#/include/explore.g
#cre.g

extern
    getChar()char,
    putChar(char ch)void,
    putString(*char message)void,
    newLine()void,
    clear()void,
    rectangle(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    line(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    pixel(ushort l, c, colour)void,
    scroll(ushort l1, c1, l2, c2; short deltaY, deltaX)void,
    installBlinker(proc()void blinker)void,
    installMouser(proc(uint yPos, xPos; bool lDown, rDown)void mouser)void,
    pointerOn()void,
    pointerOff()void,
    pointerFreeze()void,
    pointerThaw()void,
    copyIn(ushort l, c, h, w; *uint p)void,
    forceChar(char ch)void,
    graphicsDown()void,
    button(char letter; ushort l, c)void,

    putInt(int n)void,
    err(*char message)void,
    abort(*char message)void,
    getNumber(*int pVal; *char prompt;
	      int minimumValue, maximumValue; bool inCom)bool,
    getNum(*char prompt; int minimumValue, maximumValue)int,
    repNum(*char prompt; int minimumValue, maximumValue, oldValue)int,
    ask(*char question)bool,
    getBit(BITPOS pos)unsigned 1,
    putBit(BITPOS pos; unsigned 1 bit)void,
    shuffleBits(BITPOS start; uint count)void,
    noTer()void,
    noGrid()void,
    menuSelect(*uint pLastSelect; uint yPos, xPos)uint,
    setFullPalette(uint which)void,
    insertPattern(ushort row, col;
		  PATTERN tl, t, tr, l, me, r, bl, b, br)void,
    edgesOn()void,
    edgesOff()void,
    forceOn()void,
    forceOff()void,

    getFileName(*char prompt, typ)bool,
    fileOpen2()void,
    fileWrite(byte ch)void,
    fileChars(*char p)void,
    fileNum(ushort n)void,
    fileCRLF()void,
    fileClose()void;

GRIDLINE
    GridLines,			/* number of lines in current grid */
    WindowLineTop,		/* window position in current grid */
    CursorLine; 		/* cursor position in current grid */

GRIDCOL
    GridCols,			/* number of cols in current grid */
    WindowColLeft,		/* window position in current grid */
    CursorCol;			/* cursor position in current grid */

GRID
    GridCurrent;		/* number of current grid */

*GRIDTYPE
    GridPtr;			/* pointer for easy access */

uint LastSelect, CurrentTerr, TopTerr, TerrPos;

uint NO_TERR = T_SPECIAL + 1;

/*
 * sayGrid - print the grid editting prompt.
 */

proc sayGrid()void:

    putString("GRID> ");
corp;

/*
 * gridError - an error while editing a grid.
 */

proc gridError(*char message)void:

    err(message);
    sayGrid();
corp;

/*
 * gridBlink - blinker routine - we don't use blinking.
 */

proc gridBlink()void:

corp;

/*
 * getTerritory - pull a terrain value out of the BitRaw data.
 */

proc getTerritory(GRIDLINE l; GRIDCOL c)TERRITORY:
    BITPOS pos;
    uint i;
    TERRITORY t;

    pos := (make(l, BITPOS) * GridCols + c) * TERRAINBITS;
    t := 0;
    for i from 0 upto TERRAINBITS - 1 do
	t := t << 1 | getBit(pos + i);
    od;
    t
corp;

/*
 * putTerritory - stuff a terrain value into the BitRaw data.
 */

proc putTerritory(GRIDLINE l; GRIDCOL c; TERRITORY t)void:
    BITPOS pos;
    uint i;

    pos := (make(l, BITPOS) * GridCols + c) * TERRAINBITS;
    for i from TERRAINBITS - 1 downto 0 do
	putBit(pos + i, t & 1);
	t := t >> 1;
    od;
corp;

/*
 * lookUpGridSpecial - see if there is a special entry for the current GRID
 *		       at the given position.
 */

proc lookUpGridSpecial(GRIDLINE l; GRIDCOL c)*GRIDSPECIALTYPE:
    *GRIDSPECIALTYPE sp;

    sp := &GridSpecial[0];
    while sp ~= &GridSpecial[GridSpecialNext] and
	    (sp*.gsp_line ~= l or sp*.gsp_col ~= c or
	     sp*.gsp_grid ~= GridCurrent) do
	sp := sp + sizeof(GRIDSPECIALTYPE);
    od;
    if sp = &GridSpecial[GridSpecialNext] then
	nil
    else
	sp
    fi
corp;

/*
 * getGridSpecial - retrieve the SPECIAL entry which the grid says is here.
 */

proc getGridSpecial(GRIDLINE l; GRIDCOL c)*GRIDSPECIALTYPE:
    *GRIDSPECIALTYPE sp;

    sp := lookUpGridSpecial(l, c);
    if sp = nil then
	abort("special not found");
    fi;
    sp
corp;

/*
 * putGridSpecial - store a new special entry for this grid.
 */

proc putGridSpecial(GRIDLINE l; GRIDCOL c; TERRITORY t)bool:
    *GRIDSPECIALTYPE sp;

    sp := lookUpGridSpecial(l, c);
    if sp ~= nil then
	sp*.gsp_territory := t;
	true
    else
	if GridSpecialNext = GRIDSPECIALMAX then
	    gridError("Too many special grid elements");
	    forceChar('\e');
	    false
	else
	    sp := &GridSpecial[GridSpecialNext];
	    sp*.gsp_grid := GridCurrent;
	    sp*.gsp_line := l;
	    sp*.gsp_col := c;
	    sp*.gsp_territory := t;
	    GridSpecialNext := GridSpecialNext + 1;
	    true
	fi
    fi
corp;

/*
 * deleteGridSpecial - delete the current special entry here.
 */

proc deleteGridSpecial(GRIDLINE l; GRIDCOL c)void:
    *GRIDSPECIALTYPE sp1, sp2;

    sp1 := lookUpGridSpecial(l, c);
    if sp1 ~= nil then
	sp2 := sp1 + sizeof(GRIDSPECIALTYPE);
	while sp2 ~= &GridSpecial[GridSpecialNext] do
	    sp1* := sp2*;
	    sp1 := sp2;
	    sp2 := sp2 + sizeof(GRIDSPECIALTYPE);
	od;
	GridSpecialNext := GridSpecialNext - 1;
    fi;
corp;

/*
 * displayPattern - display the given pattern at the given screen position.
 */

proc displayPattern(SCREENLINE sl; SCREENCOL sc; *PATTERNTYPE pa)void:

    copyIn(sl, sc, PATTERNLINES, PATTERNCOLS, &pa*.pa_data[0, 0, 0]);
corp;

/*
 * findPattern - find the pattern to display for the given position in
 *		 the current grid.
 */

proc findPattern(GRIDLINE l; GRIDCOL c)PATTERN:
    TERRAIN t;

    t := getTerritory(l, c);
    Territory[if t = T_SPECIAL then
		  getGridSpecial(l, c)*.gsp_territory
	      else
		  GridPtr*.g_territorys[t]
	      fi].t_pattern
corp;

/*
 * displayAt - display the pattern that should be at the given screen
 *	       position in the current grid.
 */

proc displayAt(DISPLAYLINE l1; DISPLAYCOL c1)void:
    SCREENLINE l;
    SCREENCOL c;

    l := l1 + WindowLineTop;
    c := c1 + WindowColLeft;
    insertPattern(l1 * PATTERNLINES, c1 * PATTERNCOLS,
	    if l = 0 or c = 0 then
		NOPATTERN
	    else
		findPattern(l - 1, c - 1)
	    fi,
	    if l = 0 then
		NOPATTERN
	    else
		findPattern(l - 1, c)
	    fi,
	    if l = 0 or c = GridCols - 1 then
		NOPATTERN
	    else
		findPattern(l - 1, c + 1)
	    fi,
	    if c = 0 then
		NOPATTERN
	    else
		findPattern(l, c - 1)
	    fi,
	    findPattern(l, c),
	    if c = GridCols - 1 then
		NOPATTERN
	    else
		findPattern(l, c + 1)
	    fi,
	    if l = GridLines - 1 or c = 0 then
		NOPATTERN
	    else
		findPattern(l + 1, c - 1)
	    fi,
	    if l = GridLines - 1 then
		NOPATTERN
	    else
		findPattern(l + 1, c)
	    fi,
	    if l = GridLines - 1 or c = GridCols - 1 then
		NOPATTERN
	    else
		findPattern(l + 1, c + 1)
	    fi
    );
/*
    displayPattern(l * PATTERNLINES, c * PATTERNCOLS,
		   findPattern(l + WindowLineTop, c + WindowColLeft));
*/
corp;

/*
 * displayWindow - display the current screenful of patterns.
 */

proc displayWindow()void:
    DISPLAYLINE l;
    DISPLAYCOL c;

    edgesOff();
    for l from 0 upto DISPLAYLINES - 1 do
	for c from 0 upto DISPLAYCOLS - 1 do
	    displayAt(l, c);
	od;
    od;
    edgesOn();
corp;

/*
 * scrollGrids - scroll the grid region by the given amounts.
 */

proc scrollGrids(short deltaY, deltaX)void:

    scroll(0, 0, DISPLAYLINES * PATTERNLINES - 1,
	   DISPLAYCOLS * PATTERNCOLS - 1, deltaY, deltaX);
corp;

/*
 * windowUp - move the window up.
 */

proc windowUp()void:
    DISPLAYCOL c;

    if WindowLineTop ~= 0 then
	WindowLineTop := WindowLineTop - 1;
	CursorLine := WindowLineTop;
	scrollGrids(-PATTERNLINES, 0);
	for c from 0 upto DISPLAYCOLS - 1 do
	    displayAt(0, c);
	od;
    fi;
corp;

/*
 * windowDown - move the window down.
 */

proc windowDown()void:
    DISPLAYCOL c;

    if WindowLineTop ~= GridLines - DISPLAYLINES then
	WindowLineTop := WindowLineTop + 1;
	CursorLine := WindowLineTop + (DISPLAYLINES - 1);
	scrollGrids(PATTERNLINES, 0);
	for c from 0 upto DISPLAYCOLS - 1 do
	    displayAt(DISPLAYLINES - 1, c);
	od;
    fi;
corp;

/*
 * windowLeft - move the window left.
 */

proc windowLeft()void:
    DISPLAYLINE l;

    if WindowColLeft ~= 0 then
	WindowColLeft := WindowColLeft - 1;
	CursorCol := WindowColLeft;
	scrollGrids(0, -PATTERNCOLS);
	for l from 0 upto DISPLAYLINES - 1 do
	    displayAt(l, 0);
	od;
    fi;
corp;

/*
 * windowRight - move the window right.
 */

proc windowRight()void:
    DISPLAYLINE l;

    if WindowColLeft ~= GridCols - DISPLAYCOLS then
	WindowColLeft := WindowColLeft + 1;
	CursorCol := WindowColLeft + (DISPLAYCOLS - 1);
	scrollGrids(0, PATTERNCOLS);
	for l from 0 upto DISPLAYLINES - 1 do
	    displayAt(l, DISPLAYCOLS - 1);
	od;
    fi;
corp;

/*
 * newTerritory - assign the given new terrain value to the cursor position.
 */

proc newTerritory(TERRITORY t)void:

    deleteGridSpecial(CursorLine, CursorCol);
    putTerritory(CursorLine, CursorCol, t);
    forceOn();
    displayAt(CursorLine - WindowLineTop, CursorCol - WindowColLeft);
    forceOff();
corp;

/*
 * newGridSpecial - assign the given new special terrain to the current pos.
 */

proc newGridSpecial(TERRITORY t)void:

    if putGridSpecial(CursorLine, CursorCol, t) then
	putTerritory(CursorLine, CursorCol, T_SPECIAL);
	displayAt(CursorLine - WindowLineTop, CursorCol - WindowColLeft);
    fi;
corp;

/*
 * gridMap - print a map of the grid.
 */

proc gridMap()void:
    GRIDLINE l;
    GRIDCOL c;
    TERRITORY t;
    char ch;
    bool gotFile;

    pointerOff();
    installBlinker(nil);
    gotFile :=
	if getFileName("File to write to? ", "grd") then
	    installBlinker(gridBlink);
	    fileOpen2();
	    fileChars("Grid \#");
	    fileNum(GridCurrent);
	    fileChars(":\n\n");
	    true
	else
	    installBlinker(gridBlink);
	    putString("Grid \#");
	    putInt(GridCurrent);
	    putString(":\n\n");
	    false
	fi;
    for l from 0 upto GridLines - 1 do
	for c from 0 upto GridCols - 1 do
	    t := getTerritory(l, c);
	    ch :=
		if t = T_SPECIAL then
		    lookUpGridSpecial(l, c)*.gsp_territory + 'A'
		else
		    GridPtr*.g_territorys[t] + 'a'
		fi;
	    if gotFile then
		fileWrite(ch - '\e');
	    else
		putChar(ch);
	    fi;
	od;
	if gotFile then
	    fileCRLF();
	else
	    newLine();
	fi;
    od;
    if gotFile then
	fileCRLF();
	fileWrite(0x1a);
	fileClose();
    else
	newLine();
    fi;
    pointerOn();
    sayGrid();
    forceChar('\e');
corp;

/*
 * gridPrint - print the non-graphic info for the given grid.
 */

proc gridPrint(GRID g)void:
    *GRIDTYPE gp;
    uint i;

    gp := &Grid[g];
    putString("\nGrid \#");
    putInt(g);
    putString(" is ");
    putInt(gp*.g_lines);
    putString(" lines by ");
    putInt(gp*.g_cols);
    putString(" columns, using full palette \#");
    putInt(gp*.g_palette);
    putString(".\nIt uses the following basic territories:\n  ");
    for i from 0 upto range(TERRAIN) - 1 do
	putInt(gp*.g_territorys[i]);
	putChar(' ');
    od;
    newLine();
corp;

/*
 * gridEditMenu - display the initial menu of basic territorys for this grid.
 */

proc gridEditMenu()void:
    uint i;

    for i from 0 upto 6 do
	displayPattern(i * (PATTERNLINES + 2) + 2,
	    (DISPLAYCOLS + 1) * PATTERNCOLS - 1,
	    &Pattern*[Territory[GridPtr*.g_territorys[
				(TopTerr + i) % range(TERRAIN)]].t_pattern]);
    od;
corp;

/*
 * gridGadgets - display the control gadgets.
 */

proc gridGadgets()void:
    ushort C = DISPLAYCOLS * PATTERNCOLS + 1;

    button('<', 		   1, C);
    button('>', PATTERNLINES * 1 + 1, C);
    button('^', PATTERNLINES * 2 + 1, C);
    button('V', PATTERNLINES * 3 + 1, C);
    button('M', PATTERNLINES * 4 + 1, C);
    button('R', PATTERNLINES * 5 + 1, C);
    button('?', PATTERNLINES * 6 + 1, C);
    button('Q', PATTERNLINES * 7 + 1, C);
corp;

/*
 * gridBox - draw a box around the indicated menu pattern.
 */

proc gridBox(uint p)void:
    uint topLine;

    topLine := p * (PATTERNLINES + 2) + 1;
    line(topLine, PATTERNCOLS * (DISPLAYCOLS + 1) - 1,
	 topLine, PATTERNCOLS * (DISPLAYCOLS + 2) - 2,
	 true, 0);
    line(topLine, PATTERNCOLS * (DISPLAYCOLS + 2) - 1,
	 topLine + PATTERNLINES + 1, PATTERNCOLS * (DISPLAYCOLS + 2) - 1,
	 true, 0);
    line(topLine + PATTERNLINES + 1, PATTERNCOLS * (DISPLAYCOLS + 2) - 2,
	 topLine + PATTERNLINES + 1, PATTERNCOLS * (DISPLAYCOLS + 1) - 1,
	 true, 0);
    line(topLine + PATTERNLINES + 1, PATTERNCOLS * (DISPLAYCOLS + 1) - 2,
	 topLine, PATTERNCOLS * (DISPLAYCOLS + 1) - 2,
	 true, 0);
corp;

/*
 * selectTerr - select a territory as the current territory.
 */

proc selectTerr(uint p)void:

    if CurrentTerr ~= NO_TERR then
	gridBox(TerrPos);
    fi;
    gridBox(p);
    TerrPos := p;
    CurrentTerr := (TopTerr + p) % range(TERRAIN);
corp;

/*
 * setSpecial - set the current position to a special grid element.
 */

proc setSpecial()void:

    pointerFreeze();
    installBlinker(nil);
    newGridSpecial(getNum("Territory type? ", 0, TerritoryNext - 1));
    installBlinker(gridBlink);
    pointerThaw();
    sayGrid();
corp;

/*
 * gridStatus - print current status.
 */

proc gridStatus(bool fromMouse)void:
    *GRIDSPECIALTYPE p;

    gridPrint(GridCurrent);
    if not fromMouse then
	putString("Cursor is at (");
	putInt(CursorLine);
	putString(", ");
	putInt(CursorCol);
	p := lookUpGridSpecial(CursorLine, CursorCol);
	if p = nil then
	    putString(").");
	else
	    putString("), which is grid special \#");
	    putInt((p - &GridSpecial[0]) / sizeof(GRIDSPECIALTYPE));
	    putString(", territory \#");
	    putInt(p*.gsp_territory);
	    putChar('.');
	fi;
	newLine();
    fi;
    putString(
	"Commands are:\n"
	"  q - exit grid editing        ? - this info\n"
	"  m - print grid map           r - redraw window\n"
	"  <,4 - window left            >,6 - window right\n"
	"  ^,8 - window up              v,2 - window down\n"
	"  7 - window up-left           9 - window up-right\n"
	"  1 - window down-left         3 - window down-right\n"
    );
    sayGrid();
corp;

/*
 * gridMouse - handle mouse actions during grid editing.
 */

proc gridMouse(uint yPos, xPos; bool leftDown, rightDown)void:
    uint newPos;

    CursorLine := (yPos / PATTERNLINES) + WindowLineTop;
    CursorCol := (xPos / PATTERNCOLS) + WindowColLeft;
    if leftDown then
	if xPos >= (DISPLAYCOLS + 1) * PATTERNCOLS - 1 then
	    /* Cursor is over the territory palette */
	    LastSelect := 0;
	    newPos := (yPos - 1) / (PATTERNLINES + 2);
	    if newPos ~= TerrPos then
		selectTerr(newPos);
	    fi;
	elif xPos < DISPLAYCOLS * PATTERNCOLS then
	    /* Cursor is over the drawing area */
	    LastSelect := 0;
	    if CurrentTerr = NO_TERR then
		gridError("Select a territory first");
	    else
		newTerritory(CurrentTerr);
	    fi;
	else
	    /* pick on a gadget */
	    case menuSelect(&LastSelect, yPos, xPos)
	    incase 1:
		windowLeft();
	    incase 2:
		windowRight();
	    incase 3:
		windowUp();
	    incase 4:
		windowDown();
	    incase 5:
		gridMap();
		forceChar('\e');
	    incase 6:
		displayWindow();
	    incase 7:
		gridStatus(true);
		forceChar('\e');
	    incase 8:
		forceChar('q');
	    esac;
	fi;
    elif rightDown then
	LastSelect := 0;
	if xPos >= (DISPLAYCOLS + 1) * PATTERNCOLS - 1 then
	    /* Cursor is over the territory palette */
	    newPos := (yPos - 1) / (PATTERNLINES + 2) * 7;
	    if newPos ~= TopTerr then
		TopTerr := newPos;
		gridEditMenu();
		gridBox(TerrPos);
		CurrentTerr := NO_TERR;
		TerrPos := 8;
	    fi;
	elif xPos < DISPLAYCOLS * PATTERNCOLS then
	    /* Cursor is over the drawing area */
	    setSpecial();
	fi;
    else
	LastSelect := 0;
    fi;
corp;

/*
 * gridCommands - get and do commands when editing a grid.
 */

proc gridCommands()void:
    char ch;

    forceOff();
    setFullPalette(GridPtr*.g_palette);
    installBlinker(gridBlink);
    clear();
    displayWindow();
    TopTerr := 0;
    gridEditMenu();
    gridGadgets();
    CurrentTerr := NO_TERR;
    TerrPos := 8;
    LastSelect := 0;
    selectTerr(0);
    putString("Use right mouse button to set grid specials.\n");
    installMouser(gridMouse);
    pointerOn();
    graphicsDown();
    sayGrid();
    while
	ch := getChar();
	ch ~= 'q' and ch ~= 'Q' and ch ~= '\(0x03)'
    do
	case ch
	incase '\r':
	incase '\n':
	    newLine();
	    sayGrid();
	incase '4':
	incase '<':
	incase '\(0xc3)':
	    windowLeft();
	incase '6':
	incase '>':
	incase '\(0xc2)':
	    windowRight();
	incase '8':
	incase '^':
	incase '\(0xc0)':
	    windowUp();
	incase '2':
	incase 'v':
	incase 'V':
	incase '\(0xc1)':
	    windowDown();
	incase '1':
	    windowDown();
	    windowLeft();
	incase '3':
	    windowDown();
	    windowRight();
	incase '7':
	    windowUp();
	    windowLeft();
	incase '9':
	    windowUp();
	    windowRight();
	incase 'm':
	incase 'M':
	    gridMap();
	incase 'r':
	incase 'R':
	    displayWindow();
	incase '?':
	incase '\(0xfe)':
	    gridStatus(false);
	default:
	    gridError("Invalid command");
	esac;
    od;
    newLine();
    pointerOff();
    installMouser(nil);
    installBlinker(nil);
corp;

/*
 * gridSetup - set up for using the current grid.
 */

proc gridSetup(GRID g)void:

    GridCurrent := g;
    GridPtr := &Grid[GridCurrent];
    GridLines := GridPtr*.g_lines;
    WindowLineTop := (GridLines - DISPLAYLINES) / 2;
    CursorLine := WindowLineTop;
    GridCols := GridPtr*.g_cols;
    WindowColLeft := (GridCols - DISPLAYCOLS) / 2;
    CursorCol := WindowColLeft;
    BitPosStart := GridPtr*.g_start;
corp;

/*
 * getGridOther - get/replace the changeable, non-graphics stuff.
 */

proc getGridOther()void:
    *char prompt;
    uint i;

    putString("Enter the territories which are the basic "
	      "territories for this grid:\n");
    prompt := "Basic territory xx: ";
    for i from 0 upto range(TERRAIN) - 1 do
	(prompt + 16)* := i / 10 + '0';
	(prompt + 17)* := i % 10 + '0';
	GridPtr*.g_territorys[i] :=
	    repNum(prompt, 0, TerritoryNext - 1, GridPtr*.g_territorys[i]);
    od;
    GridPtr*.g_palette :=
	repNum("Full palette to use for grid? ", 0, FullPaletteNext - 1,
	       GridPtr*.g_palette);
corp;

/*
 * gridCreate - create a new grid.
 */

proc gridCreate()void:
    GRIDLINE l;
    GRIDCOL c;
    uint lines, cols, size, t;
    PATTERN p;

    if GridNext = GRIDMAX then
	err("No room for another grid");
    elif TerritoryNext = 0 then
	noTer();
    else
	lines := getNum("Number of lines in grid? ", 8, 255);
	GridPtr := &Grid[GridNext];
	GridPtr*.g_lines := lines;
	cols := getNum("Number of columns in grid? ", 8, 255);
	GridPtr*.g_cols := cols;
	t := lines * cols * TERRAINBITS;
	size := (t + 7) / 8;
	if t / lines / TERRAINBITS ~= cols or
		BitPosNext + size > BITPOSMAX then
	    err("No room for a grid that large");
	else
	    GridPtr*.g_start := BitPosNext;
	    BitPosNext := BitPosNext + size;
	    gridSetup(GridNext);
	    getGridOther();
	    for l from 0 upto GridLines - 1 do
		for c from 0 upto GridCols - 1 do
		    putTerritory(l, c, 0);
		od;
	    od;
	    putString("The grid you are creating is grid \#");
	    putInt(GridNext);
	    putString(".\n");
	    GridNext := GridNext + 1;
	    gridCommands();
	fi;
    fi;
corp;

/*
 * gridEdit - edit an existing grid.
 */

proc gridEdit()void:
    int n;

    if GridNext = 0 then
	noGrid();
    elif getNumber(&n, "Edit which grid? ", 0, GridNext - 1, true) then
	gridSetup(n);
	if ask("Replace basic territories/palette? ") then
	    Editing := true;
	    getGridOther();
	    Editing := false;
	fi;
	gridCommands();
    fi;
corp;

/*
 * gridPatternMenu - display the menu of available patterns.
 */

proc gridPatternMenu()void:
    PATTERN p;
    ushort l, c;

    clear();
    setFullPalette(LastFullPalette);
    p := 8;
    l := 0;
    c := 0;
    while p ~= PatternNext and l ~= DISPLAYLINES do
	displayPattern(l * PATTERNLINES, c * PATTERNCOLS, &Pattern*[p]);
	if c = 9 then
	    c := 0;
	    l := l + 1;
	else
	    c := c + 1;
	fi;
	p := p + 1;
    od;
    if l = DISPLAYLINES then
	err("Too many patterns to display");
    fi;
    graphicsDown();
corp;

/*
 * gridDelete - delete a selected grid from the world.
 */

proc gridDelete()void:
    *GRIDSPECIALTYPE sp1, sp2, spEnd;
    int n;
    GRID g;

    if GridNext = 0 then
	noGrid();
    elif getNumber(&n, "Delete which grid? ", 0, GridNext - 1, true) then
	g := n;
	sp1 := &GridSpecial[0];
	sp2 := sp1;
	spEnd := &GridSpecial[GridSpecialNext];
	while sp1 ~= spEnd do
	    sp2* := sp1*;
	    if sp1*.gsp_grid = g then
		GridSpecialNext := GridSpecialNext - 1;
	    else
		if sp2*.gsp_grid > g then
		    sp2*.gsp_grid := sp2*.gsp_grid - 1;
		fi;
		sp2 := sp2 + sizeof(GRIDSPECIALTYPE);
	    fi;
	    sp1 := sp1 + sizeof(GRIDSPECIALTYPE);
	od;
	shuffleBits(Grid[g].g_start,
		    (make(Grid[g].g_lines, uint) *
			  Grid[g].g_cols * TERRAINBITS + 7) / 8);
	GridNext := GridNext - 1;
	while g ~= GridNext do
	    Grid[g] := Grid[g + 1];
	    g := g + 1;
	od;
	putString("Grid deleted.\n");
    fi;
corp;
