#/include/explore.g
#cre.g

ushort
    MAZEVSIZE = 16,
    MAZEHSIZE = 16,
    MAZEVCNT = (SCREENLINES - MAZEVSIZE / 2) / MAZEVSIZE,
    MAZEHCNT = (SCREENCOLS - MAZEHSIZE / 2) / MAZEHSIZE,
    T = 3,
    M = 8,
    B = 13,
    H = B - T + 1,
    L = 3,
    C = 8,
    R = 13,
    W = R - L + 1;

extern
    putChar(char ch)void,
    putString(*char message)void,
    newLine()void,
    line(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    clear()void,
    installBlinker(proc()void blinker)void,
    graphicsDown()void,
    loadFullPalette([PATTERNCOLOURS] uint palette)void,

    putInt(uint n)void,
    err(*char message)void,
    getNumber(*int pVal; *char prompt;
	      int minimumValue, maximumValue; bool inCom)bool,
    getNum(*char prompt; int minimumValue, maximumValue)int,
    repNum(*char prompt; int minimumValue, maximumValue, oldValue)int,
    getBit(BITPOS pos)unsigned 1,
    putBit(BITPOS pos; unsigned 1 bit)void,
    shuffleBits(BITPOS start; uint count)void,
    noMon()void,
    noMaze()void,
    ask(*char question)bool,
    occupantDisplay(*OCCUPANTTYPE oc)void,
    getMon(*OCCUPANTTYPE oc)void,
    getFileName(*char prompt, typ)bool,

    fileOpen2()void,
    fileWrite(byte b)void,
    fileChars(*char p)void,
    fileNum(ushort n)void,
    fileClose()void;

MAZELINE
    MazeLines,			/* number of lines in current maze */
    WindowLineTop,		/* window position in current maze */
    CursorLine, 		/* cursor position in current maze */
    DispHeight, 		/* display height of maze */
    ConfigLine, 		/* line for modification */
    DispLine;			/* line for special display */

MAZECOL
    MazeCols,			/* number of cols in current maze */
    WindowColLeft,		/* window position in current maze */
    CursorCol,			/* cursor position in current maze */
    DispWidth,			/* display width of maze */
    ConfigCol,			/* column for modification */
    DispCol;			/* column for special display */

MAZE
    MazeCurrent;		/* number of current maze */

*MAZETYPE
    MazePtr;			/* ptr to current maze */

ushort
    Config;			/* current value for modification */

bool
    WantRedraw, 		/* need to redraw when moving cursor */
    CursorPresent;

SCREENLINE C_Line;
SCREENCOL C_Col;

/*
 * sayMaze - print the maze editing prompt.
 */

proc sayMaze()void:

    putString("MAZE> ");
corp;

/*
 * mazeError - an error while maze editing.
 */

proc mazeError(*char message)void:

    err(message);
    sayMaze();
corp;

/*
 * mazeCursor - draw cursor in maze editing.
 */

proc mazeCursor()void:

    line(C_Line - 4, C_Col - 4, C_Line + 4, C_Col + 4, true, 7);
    line(C_Line + 4, C_Col - 4, C_Line - 4, C_Col + 4, true, 7);
    CursorPresent := not CursorPresent;
corp;

/*
 * getConfig - pull a configuration value out of the BitRaw data.
 */

proc getConfig(MAZELINE l; MAZECOL c)ushort:
    BITPOS pos;
    ushort i, res;

    pos := (make(l, BITPOS) * MazeCols + c) * MAZEBITS;
    res := 0;
    for i from 0 upto MAZEBITS - 1 do
	res := (res << 1) | getBit(pos);
	pos := pos + 1;
    od;
    res
corp;

/*
 * putConfig - stuff a configuration value into the BitRaw data.
 */

proc putConfig(MAZELINE l; MAZECOL c; ushort conf)void:
    BITPOS pos;
    ushort i;

    pos := (make(l, BITPOS) * MazeCols + c) * MAZEBITS;
    for i from MAZEBITS - 1 downto 0 do
	putBit(pos, conf >> i);
	pos := pos + 1;
    od;
corp;

/*
 * setPos - find screen position of top left of cell.
 */

proc setPos(MAZELINE l; MAZECOL c)void:

    DispLine := (l - WindowLineTop / 2) * MAZEVSIZE + MAZEVSIZE / 4;
    DispCol := make(c - WindowColLeft / 2, uint) * MAZEHSIZE + MAZEHSIZE / 4;
corp;

/*
 * relLine - draw a line relative to DispLine and DispCol.
 */

proc relLine(short l1, c1, l2, c2)void:

    line(DispLine + l1, DispCol + c1, DispLine + l2, DispCol + c2, true, 5);
corp;

/*
 * hallUp - display symbol for a ladder going up.
 */

proc hallUp()void:

    relLine(T, L, B, L);
    relLine(B, L, B, R);
    relLine(B, R, T, R);
corp;

/*
 * hallDown - display symbol for a ladder going down.
 */

proc hallDown()void:

    relLine(T, R, B, R);
    relLine(B, R, B, L);
    relLine(B, L, M, L);
    relLine(M, L, M, R);
corp;

/*
 * hallSpecial - display symbol for a special hall location.
 */

proc hallSpecial()void:

    relLine(T, R, T, L);
    relLine(T, L, M, L);
    relLine(M, L, M, R);
    relLine(M, R, B, R);
    relLine(B, R, B, L);
corp;

/*
 * horWall - draw/undraw a horizontal wall.
 */

proc horWall()void:

    relLine(0, 0, 0, MAZEHSIZE);
corp;

/*
 * vertWall - draw/undraw a vertical wall.
 */

proc vertWall()void:

    relLine(0, 0, MAZEVSIZE, 0);
corp;

/*
 * horDoor - draw/undraw a horizontal door.
 */

proc horDoor()void:

    relLine(-1, L, -1, R);
    relLine(+1, L, +1, R);
corp;

/*
 * vertDoor - draw/undraw a vertical door.
 */

proc vertDoor()void:

    relLine(T, -1, B, -1);
    relLine(T, +1, B, +1);
corp;

/*
 * horSpecial - draw/undraw a horizontal special door.
 */

proc horSpecial()void:

    horDoor();
    relLine(0, L, 0, R);
corp;

/*
 * vertSpecial - draw/undraw a vertical special door.
 */

proc vertSpecial()void:

    vertDoor();
    relLine(T, 0, B, 0);
corp;

/*
 * mazeDisplay - display a window of the current maze on the screen.
 */

proc mazeDisplay()void:
    MAZELINE l, lastLine;
    MAZECOL c, lastCol;
    SCREENLINE sl;
    SCREENLINE sc;
    ushort config;

    installBlinker(nil);
    if CursorPresent then
	mazeCursor();
    fi;
    clear();
    lastLine := WindowLineTop / 2 + DispHeight;
    if lastLine >= MazeLines then
	lastLine := MazeLines;
	sl := DispHeight * MAZEVSIZE + MAZEVSIZE / 4;
	line(sl, MAZEHSIZE / 4,
	     sl, DispWidth * MAZEHSIZE + MAZEHSIZE / 4, true, 5);
    fi;
    lastCol := WindowColLeft / 2 + DispWidth;
    if lastCol >= MazeCols then
	lastCol := MazeCols;
	sc := DispWidth * MAZEHSIZE + MAZEHSIZE / 4;
	line(MAZEVSIZE / 4, sc,
	     DispHeight * MAZEVSIZE + MAZEVSIZE / 4, sc, true, 5);
    fi;
    for l from WindowLineTop / 2 upto lastLine do
	for c from WindowColLeft / 2 upto lastCol do
	    config := getConfig(l, c);
	    setPos(l, c);
	    if l ~= lastLine and c ~= lastCol then
		case (config >> 4) + M_EMPTY
		incase M_LADDER_UP:
		    hallUp();
		incase M_LADDER_DOWN:
		    hallDown();
		incase M_SPECIAL:
		    hallSpecial();
		esac;
	    fi;
	    if c ~= lastCol and l ~= MazeLines then
		case (config >> 2) & 0x03 + W_WALL
		incase W_WALL:
		    horWall();
		incase W_DOOR:
		    horWall();
		    horDoor();
		incase W_SPECIAL:
		    horWall();
		    horSpecial();
		esac;
	    fi;
	    if l ~= lastLine and c ~= MazeCols then
		case config & 0x03 + W_WALL
		incase W_WALL:
		    vertWall();
		incase W_DOOR:
		    vertWall();
		    vertDoor();
		incase W_SPECIAL:
		    vertWall();
		    vertSpecial();
		esac;
	    fi;
	od;
    od;
    installBlinker(mazeCursor);
corp;

/*
 * mazeGetChar - get a character during maze editing.
 */

proc mazeGetChar()char:
    extern
	getChar()char;
    char ch;

    C_Line := (CursorLine - WindowLineTop) * MAZEVSIZE / 2 + MAZEVSIZE / 4;
    C_Col :=
	 make(CursorCol - WindowColLeft, uint) * MAZEHSIZE / 2 + MAZEHSIZE / 4;
    ch := getChar();
    if CursorPresent then
	mazeCursor();
    fi;
    ch
corp;

/*
 * decLine - try to decrement the line the cursor is on.
 */

proc decLine()void:

    if CursorLine = WindowLineTop then
	if WindowLineTop ~= 0 then
	    WindowLineTop := WindowLineTop - 2;
	    CursorLine := CursorLine - 1;
	    WantRedraw := true;
	fi;
    else
	CursorLine := CursorLine - 1;
    fi;
corp;

/*
 * incLine - try to increment the line the cursor is on.
 */

proc incLine()void:

    if CursorLine = WindowLineTop + DispHeight * 2 then
	if WindowLineTop ~= (MazeLines - DispHeight) * 2 then
	    WindowLineTop := WindowLineTop + 2;
	    CursorLine := CursorLine + 1;
	    WantRedraw := true;
	fi;
    else
	CursorLine := CursorLine + 1;
    fi;
corp;

/*
 * decCol - try to decrement the column the cursor is on.
 */

proc decCol()void:

    if CursorCol = WindowColLeft then
	if WindowColLeft ~= 0 then
	    WindowColLeft := WindowColLeft - 2;
	    CursorCol := CursorCol - 1;
	    WantRedraw := true;
	fi;
    else
	CursorCol := CursorCol - 1;
    fi;
corp;

/*
 * incCol - try to increment the column the cursor is on.
 */

proc incCol()void:

    if CursorCol = WindowColLeft + DispWidth * 2 then
	if WindowColLeft ~= (MazeCols - DispWidth) * 2 then
	    WindowColLeft := WindowColLeft + 2;
	    CursorCol := CursorCol + 1;
	    WantRedraw := true;
	fi;
    else
	CursorCol := CursorCol + 1;
    fi;
corp;

/*
 * mazeMove - move the cursor during maze editing - maybe window.
 */

proc mazeMove(char ch)void:

    WantRedraw := false;
    case ch
    incase '1':
	incLine();
	decCol();
    incase '2':
    incase '\(0xc1)':
	incLine();
    incase '3':
	incLine();
	incCol();
    incase '4':
    incase '\(0xc3)':
	decCol();
    incase '6':
    incase '\(0xc2)':
	incCol();
    incase '7':
	decLine();
	decCol();
    incase '8':
    incase '\(0xc0)':
	decLine();
    incase '9':
	decLine();
	incCol();
    esac;
    if WantRedraw then
	mazeDisplay();
    fi;
corp;

/*
 * configPut - add the value to the current config, and update.
 */

proc configPut(ushort addition)void:

    putConfig(ConfigLine, ConfigCol, Config | addition);
corp;

/*
 * hallReset - reset the value for a hall, and compute global values.
 */

proc hallReset()void:

    ConfigLine := CursorLine / 2;
    ConfigCol := CursorCol / 2;
    setPos(ConfigLine, ConfigCol);
    Config := getConfig(ConfigLine, ConfigCol);
    case (Config >> 4) + M_EMPTY
    incase M_LADDER_UP:
	hallUp();
    incase M_LADDER_DOWN:
	hallDown();
    incase M_SPECIAL:
	hallSpecial();
    esac;
    Config := Config & 0x0f;
corp;

/*
 * wallReset - reset the value for a wall, and compute global values.
 */

proc wallReset()void:

    ConfigLine := CursorLine / 2;
    ConfigCol := CursorCol / 2;
    setPos(ConfigLine, ConfigCol);
    Config := getConfig(ConfigLine, ConfigCol);
    if CursorLine & 0x01 = 0 then
	case (Config >> 2) & 0x03 + W_WALL
	incase W_EMPTY:
	    horWall();
	incase W_DOOR:
	    horDoor();
	incase W_SPECIAL:
	    horSpecial();
	esac;
	Config := Config & 0x33;
    else
	case Config & 0x03 + W_WALL
	incase W_EMPTY:
	    vertWall();
	incase W_DOOR:
	    vertDoor();
	incase W_SPECIAL:
	    vertSpecial();
	esac;
	Config := Config & 0x3c;
    fi;
corp;

/*
 * isHall - current position is in a hall.
 */

proc isHall()bool:

    CursorLine & CursorCol & 0x01 ~= 0
corp;

/*
 * checkHall - check for and disallow building in a wall.
 */

proc checkHall()bool:

    if not isHall() then
	mazeError("Not in wall");
	false
    else
	true
    fi
corp;

/*
 * mazeUp - execute the 'u' command.
 */

proc mazeUp()void:

    if checkHall() then
	hallReset();
	configPut((M_LADDER_UP - M_EMPTY) << 4);
	hallUp();
    fi;
corp;

/*
 * mazeDown - execute the 'd' command.
 */

proc mazeDown()void:

    if checkHall() then
	hallReset();
	configPut((M_LADDER_DOWN - M_EMPTY) << 4);
	hallDown();
    fi;
corp;

/*
 * checkCorner - check for and disallow building in corners.
 */

proc checkCorner()bool:

    if (CursorLine | CursorCol) & 0x01 = 0 then
	mazeError("Not in corner");
	false
    else
	true
    fi
corp;

/*
 * checkBound - check for and disallow building past edges.
 */

proc checkBound()bool:

    if CursorLine = MazeLines * 2 or CursorCol = MazeCols * 2 then
	mazeError("Not in boundary");
	false
    else
	true
    fi
corp;

/*
 * checkEdge - check for and disallow building in edges.
 */

proc checkEdge()bool:

    if CursorLine = 0 or CursorCol = 0 then
	mazeError("Not in boundary");
	false
    else
	checkBound()
    fi
corp;

/*
 * mazeHall - execute the 'h' command.
 */

proc mazeHall()void:

    if isHall() then
	hallReset();
	configPut(W_EMPTY - W_EMPTY);
    elif checkCorner() and checkEdge() then
	wallReset();
	configPut(if CursorLine & 0x01 = 0 then
		      horWall();
		      (W_EMPTY - W_WALL) << 2
		  else
		      vertWall();
		      W_EMPTY - W_WALL
		  fi);
    fi;
corp;

/*
 * mazeSpecial - execute the 's' command.
 */

proc mazeSpecial()void:

    if isHall() then
	hallReset();
	configPut((M_SPECIAL - M_EMPTY) << 4);
	hallSpecial();
    elif checkCorner() and checkBound() then
	wallReset();
	configPut(if CursorLine & 0x01 = 0 then
		      horSpecial();
		      (W_SPECIAL - W_WALL) << 2
		  else
		      vertSpecial();
		      W_SPECIAL - W_WALL
		  fi);
    fi;
corp;

/*
 * checkWall - make sure that location is in a wall.
 */

proc checkWall()bool:

    if isHall() then
	mazeError("Not in hall");
	false
    else
	checkCorner()
    fi
corp;

/*
 * mazeDoor - execute the 'o' command.
 */

proc mazeDoor()void:

    if checkWall() and checkEdge() then
	wallReset();
	configPut(if CursorLine & 0x01 = 0 then
		      horDoor();
		      (W_DOOR - W_WALL) << 2
		  else
		      vertDoor();
		      W_DOOR - W_WALL
		  fi);
    fi;
corp;

/*
 * mazeWall - execute the 'w' command.
 */

proc mazeWall()void:

    if checkWall() and checkBound() then
	wallReset();
	configPut(W_WALL - W_WALL);
    fi;
corp;

/*
 * mazePrint - print the non-graphic info for the given maze.
 */

proc mazePrint(MAZE m)void:
    *MAZETYPE ma;

    ma := &Maze[m];
    putString("\nMaze \#");
    putInt(m);
    putString(" is ");
    putInt(ma*.ma_lines);
    putString(" lines by ");
    putInt(ma*.ma_cols);
    putString(" columns.");
    occupantDisplay(&ma*.ma_occupants[0]);
corp;

/*
 * mazeMap - print a picture of the maze.
 */

proc mazeMap()void:
    MAZELINE l;
    MAZECOL c;
    ushort i, j;
    char ch;
    bool gotFile;

    installBlinker(nil);
    if CursorPresent then
	mazeCursor();
    fi;
    gotFile :=
	if getFileName("File to write to? ", "maz") then
	    fileOpen2();
	    fileChars("Maze \#");
	    fileNum(MazeCurrent);
	    fileChars(":\n\n");
	    true
	else
	    putString("Maze \#");
	    putInt(MazeCurrent);
	    putString(":\n\n");
	    false
	fi;
    for l from 0 upto MazeLines - 1 do
	for c from 0 upto MazeCols - 1 do
	    if gotFile then
		fileWrite('+' - '\e');
	    else
		putChar('+');
	    fi;
	    i := (getConfig(l, c) >> 2) & 0x03;
	    ch :=
		if i + W_WALL = W_EMPTY then
		    ' '
		elif i + W_WALL = W_DOOR then
		    '/'
		elif i + W_WALL = W_SPECIAL then
		    's'
		else
		    '-'
		fi;
	    if gotFile then
		fileWrite(ch - '\e');
	    else
		putChar(ch);
	    fi;
	od;
	if gotFile then
	    fileChars("+\n");
	else
	    putString("+\n");
	fi;
	for c from 0 upto MazeCols - 1 do
	    i := getConfig(l, c);
	    j := i & 0x03;
	    ch :=
		if j + W_WALL = W_EMPTY then
		    ' '
		elif j + W_WALL = W_DOOR then
		    '/'
		elif j + W_WALL = W_SPECIAL then
		    's'
		else
		    '|'
		fi;
	    if gotFile then
		fileWrite(ch - '\e');
	    else
		putChar(ch);
	    fi;
	    j := (i >> 4) & 0x03;
	    ch :=
		if j + M_EMPTY = M_LADDER_UP then
		    'U'
		elif j + M_EMPTY = M_LADDER_DOWN then
		    'D'
		elif j + M_EMPTY = M_SPECIAL then
		    'S'
		else
		    ' '
		fi;
	    if gotFile then
		fileWrite(ch - '\e');
	    else
		putChar(ch);
	    fi;
	od;
	if gotFile then
	    fileChars("|\n");
	else
	    putString("|\n");
	fi;
    od;
    if gotFile then
	fileWrite('+' - '\e');
    else
	putChar('+');
    fi;
    for c from 0 upto MazeCols -1 do
	if gotFile then
	    fileWrite('-' - '\e');
	    fileWrite('+' - '\e');
	else
	    putChar('-');
	    putChar('+');
	fi;
    od;
    if gotFile then
	fileChars("\n\n");
	fileClose();
    else
	putString("\n\n");
    fi;
    installBlinker(mazeCursor);
corp;

/*
 * mazeCommands - get and do commands when editing a maze.
 */

proc mazeCommands()void:
    char ch;

    clear();
    loadFullPalette(FullPalette[LastFullPalette].fp_palette);
    installBlinker(mazeCursor);
    CursorPresent := false;
    mazeDisplay();
    graphicsDown();
    sayMaze();
    while
	ch := mazeGetChar();
	ch ~= 'q' and ch ~= 'Q' and ch ~= '\(0x03)'
    do
	case ch
	incase '\r':
	incase '\n':
	    newLine();
	    sayMaze();
	incase '?':
	incase '\(0xfe)':
	    mazePrint(MazeCurrent);
	    putString("Cursor is at (");
	    putInt(CursorLine / 2);
	    putString(", ");
	    putInt(CursorCol / 2);
	    putString(").\n");
	    putString(
    "Commands are:\n"
    "  q - exit maze editing            ? - this info\n"
    "  m - print maze map               5 - window to a maze corner\n"
    "  u - build ladder going up        d - build ladder going down\n"
    "  w - build wall here              o - build door here\n"
    "  h - build empty hall here        s - put special here\n"
    "  1-4,6-9 - move cursor in corresponding direction\n"
	    );
	    sayMaze();
	incase '1':
	incase '2':
	incase '3':
	incase '4':
	incase '6':
	incase '7':
	incase '8':
	incase '9':
	incase '\(0xc0)':
	incase '\(0xc1)':
	incase '\(0xc2)':
	incase '\(0xc3)':
	    mazeMove(ch);
	incase 'c':
	incase 'C':
	incase '5':
	    putString("Window to which corner? ");
	    ch := mazeGetChar();
	    newLine();
	    case ch
	    incase '1':
		WindowLineTop := (MazeLines - DispHeight) * 2;
		CursorLine := MazeLines * 2;
		WindowColLeft := 0;
		CursorCol := 0;
	    incase '3':
		WindowLineTop := (MazeLines - DispHeight) * 2;
		CursorLine := MazeLines * 2;
		WindowColLeft := (MazeCols - DispWidth) * 2;
		CursorCol := MazeCols * 2;
	    incase '7':
		WindowLineTop := 0;
		CursorLine := 0;
		WindowColLeft := 0;
		CursorCol := 0;
	    incase '9':
		WindowLineTop := 0;
		CursorLine := 0;
		WindowColLeft := (MazeCols - DispWidth) * 2;
		CursorCol := MazeCols * 2;
	    default:
		putString("Use 1, 3, 7, or 9.\n");
	    esac;
	    mazeDisplay();
	    sayMaze();
	incase 'u':
	incase 'U':
	    mazeUp();
	incase 'd':
	incase 'D':
	    mazeDown();
	incase 'h':
	incase 'H':
	    mazeHall();
	incase 's':
	incase 'S':
	    mazeSpecial();
	incase 'o':
	incase 'O':
	    mazeDoor();
	incase 'w':
	incase 'W':
	    mazeWall();
	incase 'm':
	incase 'M':
	    mazeMap();
	    sayMaze();
	default:
	    mazeError("Invalid command; ? for help");
	esac;
    od;
    newLine();
    installBlinker(nil);
corp;

/*
 * mazeSetup - set up for using the current maze.
 */

proc mazeSetup(MAZE m)void:

    MazeCurrent := m;
    MazePtr := &Maze[MazeCurrent];
    MazeLines := MazePtr*.ma_lines;
    DispHeight := if MazeLines < MAZEVCNT then MazeLines else MAZEVCNT fi;
    WindowLineTop := 0;
    CursorLine := 0;
    MazeCols := MazePtr*.ma_cols;
    DispWidth := if MazeCols < MAZEHCNT then MazeCols else MAZEHCNT fi;
    WindowColLeft := 0;
    CursorCol := 0;
    BitPosStart := MazePtr*.ma_start;
corp;

/*
 * getMazeOther - get/replace changeable non-graphic stuff.
 */

proc getMazeOther()void:

    MazePtr*.ma_palette := repNum("Enter half palette to use with maze: ",
				  0, HalfPaletteNext - 1, MazePtr*.ma_palette);
    getMon(&MazePtr*.ma_occupants[0]);
corp;

/*
 * mazeCreate - create a new maze.
 */

proc mazeCreate()void:
    MAZELINE r;
    MAZECOL c;
    uint lines, cols, size, t;

    if MazeNext = MAZEMAX then
	err("No room for another maze");
    elif MonsterNext = 0 then
	noMon();
    else
	lines := getNum("Number of lines in maze? ", 1, 255);
	Maze[MazeNext].ma_lines := lines;
	cols := getNum("Number of columns in maze? ", 1, 255);
	Maze[MazeNext].ma_cols := cols;
	t := lines * cols * MAZEBITS;
	size := (t + 7) / 8;
	if t / lines / MAZEBITS ~= cols or
		BitPosNext + size > BITPOSMAX then
	    err("No room for a maze that large");
	else
	    Maze[MazeNext].ma_start := BitPosNext;
	    BitPosNext := BitPosNext + size;
	    mazeSetup(MazeNext);
	    getMazeOther();
	    for r from 0 upto MazeLines - 1 do
		for c from 0 upto MazeCols - 1 do
		    putConfig(r, c, 0);
		od;
	    od;
	    putString("The maze you are creating is maze \#");
	    putInt(MazeNext);
	    putString(".\n\n");
	    MazeNext := MazeNext + 1;
	    mazeCommands();
	fi;
    fi;
corp;

/*
 * mazeEdit - edit an existing maze.
 */

proc mazeEdit()void:
    int n;

    if MazeNext = 0 then
	noMaze();
    elif getNumber(&n, "Edit which maze? ", 0, MazeNext - 1, true) then
	mazeSetup(n);
	if ask("Replace occupant monsters/palette? ") then
	    Editing := true;
	    getMazeOther();
	    Editing := false;
	fi;
	mazeCommands();
    fi;
corp;

/*
 * mazeDelete - delete a selected maze from the world.
 */

proc mazeDelete()void:
    int n;
    MAZE m;

    if MazeNext = 0 then
	noMaze();
    elif getNumber(&n, "Delete which maze? ", 0, MazeNext - 1, true) then
	m := n;
	shuffleBits(Maze[m].ma_start,
		    (make(Maze[m].ma_lines, uint) *
			  Maze[m].ma_cols * MAZEBITS + 7) / 8);
	MazeNext := MazeNext - 1;
	while m ~= MazeNext do
	    Maze[m] := Maze[m + 1];
	    m := m + 1;
	od;
	putString("Maze deleted.\n");
    fi;
corp;
