#/include/explore.g
#cre.g

extern
    getChar()char,
    putChar(char ch)void,
    putString(*char message)void,
    newLine()void,
    forceChar(char ch)void,
    installBlinker(proc()void blinker)void,
    installMouser(proc(uint yPos, xPos; bool lDown, rDown)void mouser)void,
    clear()void,
    line(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    rectangle(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    pointerOn()void,
    pointerOff()void,
    loadFullPalette(FULLPALETTEARRAY palette)void,
    loadTopPalette(HALFPALETTEARRAY palette)void,
    loadColour(uint which, colour)void,
    graphicsDown()void,
    button(char letter; ushort l, c)void,
    setColourCycle(ushort which, first, last, period)void,

    putInt(uint n)void,
    err(*char message)void,
    getNumber(*int pn; *char prompt; int minV, maxV; bool inCmd)bool,
    getNum(*char prompt; int minimumValue, maximumValue)int,
    scan(*char n1, n2; *ushort p; ushort x, cnt, size)ushort,
    decr(*ushort p; ushort x, cnt, size)void,
    getFileName(*char prompt, typ)bool,
    menuSelect(*uint pLastSelect; uint yPos, xPos)uint,
    setFullPalette(uint which)void,
    setTopPalette(uint which)void,

    fileOpen(bool isRead)bool,
    fileOpen2()void,
    fileClose()void,
    fileGet(*byte buffer; uint length)void,
    filePut(*byte buffer; uint length)void;

HALFPALETTEARRAY DEFAULT_HALF_PALETTE = (
    0x2c0, /*  0 - dark green */
    0xff0, /*  1 - lemon yellow */
    0x00f, /*  2 - blue */
    0xd00, /*  3 - brick red */
    0xf90, /*  4 - orange */
    0xc80, /*  5 - brown */
    0x91f, /*  6 - purple */
    0x000, /*  7 - black */
    0xfff, /*  8 - white */
    0x0db, /*  9 - aqua */
    0xdb9, /* 10 - tan */
    0xfb0, /* 11 - golden orange */
    0xc1f, /* 12 - violet */
    0x0b1, /* 13 - forest green */
    0x6ce, /* 14 - light blue */
    0xf00  /* 15 - red */
);

FULLPALETTEARRAY DEFAULT_FULL_PALETTE = (
    0x06d, /*  0 - dark blue */
    0x000, /*  1 - black */
    0x2c0, /*  2 - dark green */
    0xff0, /*  3 - lemon yellow */
    0x00f, /*  4 - blue */
    0xd00, /*  5 - brick red */
    0xf90, /*  6 - orange */
    0xc80, /*  7 - brown */
    0x91f, /*  8 - purple */
    0xfac, /*  9 - pink */
    0x6fe, /* 10 - sky blue */
    0xa87, /* 11 - dark brown */
    0x0bb, /* 12 - blue green */
    0x0f0, /* 13 - green */
    0xf9a, /* 14 - cherry red */
    0x0db, /* 15 - aqua */
    0xdb9, /* 16 - tan */
    0xfb0, /* 17 - golden orange */
    0xc1f, /* 18 - violet */
    0x0b1, /* 19 - forest green */
    0x6ce, /* 20 - light blue */
    0xf00, /* 21 - red */
    0xfff, /* 22 - unused 0 */
    0x000, /* 23 - black */
    0x222, /* 24 - darkest grey */
    0x444, /* 25 */
    0x666, /* 26 */
    0x888, /* 27 - medium grey */
    0xaaa, /* 28 */
    0xccc, /* 29 - light grey */
    0xeee, /* 30 - lightest grey */
    0xfff  /* 31 - white */
);

ushort
    NOHIGHLIGHT = PATTERNCOLOURS;

ushort
    TOP = 5,					/* top of slider set */
    LEFT = 0,					/* left of slider set */
    SCALE = 6,
    VERT_SLACK = 1,				/* space before knob */
    TOP_START = TOP + VERT_SLACK,
    HORIZ_GAP = 5,				/*half of gap between sliders*/
    KNOB_EXTRA = 2,				/* knob sticks out */
    BODY_WIDTH = 7,
    KNOB_WIDTH = BODY_WIDTH + KNOB_EXTRA * 2,
    KNOB_HEIGHT = KNOB_WIDTH,
    TOTAL_WIDTH = HORIZ_GAP * 2 + KNOB_WIDTH,
    RED_START = LEFT + HORIZ_GAP,
    GREEN_START = RED_START + TOTAL_WIDTH,
    BLUE_START = GREEN_START + TOTAL_WIDTH,
    BOTTOM = TOP + VERT_SLACK * 2 + KNOB_HEIGHT + SCALE * 15;

HALFPALETTE CurrentHalf;
FULLPALETTE CurrentFull;
*HALFPALETTETYPE HalfPtr;
*FULLPALETTETYPE FullPtr;
bool IsHalf;

uint LastSelect;

ushort CurrentColour, HighLightColour, RedPos, GreenPos, BluePos, KnobOffset;

enum {Sel_none, Sel_red, Sel_green, Sel_blue} SelectState;

enum {Cy_none, Cy_first, Cy_last} CycleState;
*CYCLETYPE CyclePtr;
ushort CurrentCycle;

/*
 * sayPal - print the palette editing prompt.
 */

proc sayPal()void:

    putString("PALETTE> ");
corp;

/*
 * palErr - issue an error message during palette editing.
 */

proc palErr(*char message)void:

    err(message);
    sayPal();
corp;

/*
 * palBlink - blinker routine - we don't use blinking.
 */

proc palBlink()void:

corp;

/*
 * palScales - draw scales for the sliders.
 */

proc palScales()void:
    ushort i, row;

    for i from 0 upto 15 do
	row := TOP_START + KNOB_HEIGHT / 2 + i * SCALE;
	line(row, LEFT + 1, row, LEFT + TOTAL_WIDTH * 3 - 1, false, 1);
    od;
corp;

/*
 * getHexChar - get a single hexadecimal character value.
 */

proc getHexChar(uint i)char:

    if i > 9 then
	i - 10 + 'A'
    else
	i + '0'
    fi
corp;

/*
 * reDrawButton - redraw the digit button.
 */

proc reDrawButton(ushort column, valu)void:

    rectangle(BOTTOM + 5, column + 1, BOTTOM + 13, column + 9, false, 0);
    button(getHexChar(valu), BOTTOM + 3, column - 1);
corp;

/*
 * reDrawKnob - knob has been moved - redraw it.
 */

proc reDrawKnob(ushort column, oldPos, newPos)void:

    /* first, undraw the first - draw in background, then slider */
    rectangle(TOP_START + oldPos, column,
	      (TOP_START + KNOB_HEIGHT - 1) + oldPos, column + KNOB_WIDTH,
	      false, 0);
    rectangle(TOP_START + oldPos, column + KNOB_EXTRA,
	      (TOP_START + KNOB_HEIGHT - 1) + oldPos,
		column + (BODY_WIDTH + KNOB_EXTRA),
	      false, 1);
    /* now put in the new knob */
    rectangle(TOP_START + newPos, column,
	      (TOP_START + KNOB_HEIGHT - 1) + newPos, column + KNOB_WIDTH,
	      false, 2);
corp;

/*
 * palKnobs - redraw all knobs.
 */

proc palKnobs(uint colour)void:
    ushort red, green, blue;
    uint redPos, greenPos, bluePos;

    red := (colour >> 8) & 0xf;
    green := (colour >> 4) & 0xf;
    blue := colour & 0xf;
    redPos := (15 - red) * SCALE;
    greenPos := (15 - green) * SCALE;
    bluePos := (15 - blue) * SCALE;
    reDrawKnob(RED_START, RedPos, redPos);
    reDrawKnob(GREEN_START, GreenPos, greenPos);
    reDrawKnob(BLUE_START, BluePos, bluePos);
    RedPos := redPos;
    GreenPos := greenPos;
    BluePos := bluePos;
    reDrawButton(RED_START, red);
    reDrawButton(GREEN_START, green);
    reDrawButton(BLUE_START, blue);
corp;

/*
 * drawSlider - draw one slider at the given column with the given value.
 */

proc drawSlider(ushort column, pos, valu)void:

    /* first the backgound (erases part of the scales lines) */
    rectangle(TOP, column, BOTTOM, column + KNOB_WIDTH, false, 0);
    /* now the slider column */
    rectangle(TOP, column + KNOB_EXTRA,
	      BOTTOM - 1, column + (BODY_WIDTH + KNOB_EXTRA), false, 1);
    /* and the knob itself */
    rectangle(TOP_START + pos, column,
	      (TOP_START + KNOB_HEIGHT - 1) + pos, column + KNOB_WIDTH,
	      false, 2);
    button(getHexChar(valu), BOTTOM + 3, column - 1);
corp;

/*
 * palSliders - draw the slider gadgets.
 */

proc palSliders()void:
    uint colour;
    ushort red, green, blue;

    colour :=
	if IsHalf then
	    HalfPtr*.hp_palette[CurrentColour]
	else
	    FullPtr*.fp_palette[CurrentColour]
	fi;
    red := (colour >> 8) & 0xf;
    green := (colour >> 4) & 0xf;
    blue := colour & 0xf;
    RedPos := (15 - red) * SCALE;
    GreenPos := (15 - green) * SCALE;
    BluePos := (15 - blue) * SCALE;
    drawSlider(RED_START, RedPos, red);
    drawSlider(GREEN_START, GreenPos, green);
    drawSlider(BLUE_START, BluePos, blue);
corp;

/*
 * paletteGadgets - display the palette editing gadgets.
 */

proc paletteGadgets()void:
    ushort C = DISPLAYCOLS * PATTERNCOLS + 1;

    button(' ', 		   1, C);	/* 'I/O' */
    line( 4, C +  3,  4, C +  5, false, 2);
    line( 4, C +  4,  7, C +  4, false, 2);
    line( 7, C +  3,  7, C +  5, false, 2);
    line( 4, C + 10, 11, C +  3, false, 2);
    line( 8, C +  8,  8, C + 10, false, 2);
    line( 8, C + 10, 11, C + 10, false, 2);
    line(11, C + 10, 11, C +  8, false, 2);
    line(11, C +  8,  8, C +  8, false, 2);
    button('R', PATTERNLINES * 1 + 1, C);
    button('?', PATTERNLINES * 2 + 1, C);
    button('Q', PATTERNLINES * 3 + 1, C);
    button('0', PATTERNLINES * 4 + 1, C);
    button('1', PATTERNLINES * 5 + 1, C);
    if not IsHalf then
	button('2', PATTERNLINES * 6 + 1, C);
	button('3', PATTERNLINES * 7 + 1, C);
    fi;

    palScales();
    palSliders();
corp;

/*
 * paletteRead - read the current palette from a .hlf or .ful file.
 */

proc paletteRead()void:

    installBlinker(nil);
    pointerOff();
    if IsHalf then
	if getFileName("File to read from? ", "hlf") then
	    if not fileOpen(true) then
		palErr("Palette file doesn't exist");
	    else
		fileGet(pretend(HalfPtr, *byte), sizeof(HALFPALETTETYPE));
		fileClose();
		palKnobs(HalfPtr*.hp_palette[CurrentColour]);
		sayPal();
	    fi;
	else
	    putString("OK, no file read.\n");
	    sayPal();
	fi;
    else
	if getFileName("File to read from? ", "ful") then
	    if not fileOpen(true) then
		palErr("Palette file doesn't exist");
	    else
		fileGet(pretend(FullPtr, *byte), sizeof(FULLPALETTETYPE));
		fileClose();
		palKnobs(FullPtr*.fp_palette[CurrentColour]);
		sayPal();
	    fi;
	else
	    putString("OK, no file read.\n");
	    sayPal();
	fi;
    fi;
    pointerOn();
    installBlinker(palBlink);
    forceChar('\e');
corp;

/*
 * paletteWrite - write the current palette to a .hlf or .ful file.
 */

proc paletteWrite()void:

    installBlinker(nil);
    pointerOff();
    if IsHalf then
	if getFileName("File to write to? ", "hlf") then
	    fileOpen2();
	    filePut(pretend(HalfPtr, *byte), sizeof(HALFPALETTETYPE));
	    fileClose();
	else
	    putString("OK, no file written.\n");
	fi;
    else
	if getFileName("File to write to? ", "ful") then
	    fileOpen2();
	    filePut(pretend(FullPtr, *byte), sizeof(FULLPALETTETYPE));
	    fileClose();
	else
	    putString("OK, no file written.\n");
	fi;
    fi;
    pointerOn();
    installBlinker(palBlink);
    sayPal();
    forceChar('\e');
corp;

/*
 * paletteReset - reset the current palette to the default colours.
 */

proc paletteReset()void:

    if IsHalf then
	HalfPtr*.hp_palette := DEFAULT_HALF_PALETTE;
	loadTopPalette(HalfPtr*.hp_palette);
	palKnobs(HalfPtr*.hp_palette[CurrentColour]);
    else
	FullPtr*.fp_palette := DEFAULT_FULL_PALETTE;
	loadFullPalette(FullPtr*.fp_palette);
	palKnobs(FullPtr*.fp_palette[CurrentColour]);
    fi;
corp;

/*
 * paletteHelp - help during palette editing.
 */

proc paletteHelp()void:
    uint i;

    if IsHalf then
	putString("\nEditing half palette \#");
	putInt(CurrentHalf);
	i := HalfPtr*.hp_palette[CurrentColour];
    else
	putString("\nEditing full palette \#");
	putInt(CurrentFull);
	i := FullPtr*.fp_palette[CurrentColour];
    fi;
    putString(", current colour is \#");
    putInt(CurrentColour);
    putString(" which has RGB values ");
    putChar(getHexChar(i >> 8));
    putChar(' ');
    putChar(getHexChar((i >> 4) & 0xf));
    putChar(' ');
    putChar(getHexChar(i & 0xf));
    putString(
	"\nCommands are:\n"
	"  q - exit palette editing     ? - this info\n"
	"  i - input palette from file  o - output palette to file\n"
	"  r - reset to default colours\n"
	"  0 - define colour cycle #0   1 - define colour cycle #1\n"
	"  2 - define colour cycle #2   3 - define colour cycle #3\n"
    );
    sayPal();
corp;

/*
 * palColourBox - XOR draw a box on the given colour's menu box.
 */

proc palColourBox(ushort colour)void:
    uint xPos, yPos;

    if IsHalf then
	xPos := 160 - 16 + colour /  8 * 8;
	yPos := colour % 8 * 8;
	rectangle(yPos, xPos, yPos + 7, xPos + 7, true, PICTURECOLOURS - 1);
    else
	xPos := 160 - 16 + colour / 16 * 8;
	yPos := colour % 16 * 8;
	rectangle(yPos, xPos, yPos + 7, xPos + 7, true, PATTERNCOLOURS - 1);
    fi;
corp;

/*
 * palUnHighLight - remove any highlight of a colour menu box.
 */

proc palUnHighLight()void:

    if HighLightColour ~= NOHIGHLIGHT then
	palColourBox(HighLightColour);
	HighLightColour := NOHIGHLIGHT;
    fi;
corp;

/*
 * palMainBox - draw the main colour box.
 */

proc palMainBox()void:

    rectangle(10, 74, 54, 118, false,
	      if IsHalf then CurrentColour + 16 else CurrentColour fi);
corp;

/*
 * startCycle - start defining a colour cycle.
 */

proc startCycle(uint which)void:

    if IsHalf and which > 1 then
	palErr("Only cycles 0 and 1 available on half palettes");
    else
	CycleState := Cy_first;
	CyclePtr :=
	    if IsHalf then
		&HalfPtr*.hp_cycles[which]
	    else
		&FullPtr*.fp_cycles[which]
	    fi;
	CurrentCycle := which;
	putString("\nSelect first colour in cycle: ");
    fi;
    forceChar('\e');
corp;

/*
 * palMouse - handle mouse events for palette editing.
 */

proc palMouse(uint yPos, xPos; bool leftDown, rightDown)void:
    uint oldColour, newColour, newValue, pos;

    if leftDown then
	if xPos >= 160 - 16 then
	    LastSelect := 0;
	    /* Cursor is over the colour menu */
	    if IsHalf then
		newColour := (xPos - (160 - 16)) / 8 *	8 + yPos / 8;
	    else
		newColour := (xPos - (160 - 16)) / 8 * 16 + yPos / 8;
	    fi;
	    if newColour ~= HighLightColour then
		palUnHighLight();
		HighLightColour := newColour;
		CurrentColour := newColour;
		palColourBox(newColour);
		palMainBox();
		palKnobs(
		    if IsHalf then
			HalfPtr*.hp_palette[CurrentColour]
		    else
			FullPtr*.fp_palette[CurrentColour]
		    fi);
		if CycleState = Cy_first then
		    CyclePtr*.cy_firstColour := newColour;
		    CycleState := Cy_last;
		    putString("\nSelect last colour in cycle: ");
		    forceChar('\e');
		    palUnHighLight();
		elif CycleState = Cy_last then
		    if newColour < CyclePtr*.cy_firstColour then
			newLine();
			palErr("Last colour must be >= first colour");
			CycleState := Cy_none;
		    else
			CyclePtr*.cy_lastColour := newColour;
			CycleState := Cy_none;
			newLine();
			installBlinker(nil);
			pointerOff();
			CyclePtr*.cy_period :=
			    getNum("Enter cycle period (0 - 255): ", 0, 255);
			pointerOn();
			installBlinker(palBlink);
			if IsHalf then
			    setColourCycle(CurrentCycle + 2,
				CyclePtr*.cy_firstColour + PICTURECOLOURS,
				newColour + PICTURECOLOURS,
				CyclePtr*.cy_period);
			else
			    setColourCycle(CurrentCycle,
				CyclePtr*.cy_firstColour, newColour,
				CyclePtr*.cy_period);
			fi;
			sayPal();
		    fi;
		    forceChar('\e');
		    palUnHighLight();
		fi;
	    fi;
	elif xPos < DISPLAYCOLS * PATTERNCOLS then
	    palUnHighLight();
	    LastSelect := 0;
	    if CycleState ~= Cy_none then
		;
	    elif SelectState = Sel_none then
		if yPos >= TOP_START + RedPos and
			yPos < TOP_START + KNOB_HEIGHT + RedPos and
			xPos >= RED_START and
			xPos < RED_START + KNOB_WIDTH then
		    SelectState := Sel_red;
		    KnobOffset := yPos - (TOP_START + RedPos);
		elif yPos >= TOP_START + GreenPos and
			yPos < TOP_START + KNOB_HEIGHT + GreenPos and
			xPos >= GREEN_START and
			xPos < GREEN_START + KNOB_WIDTH then
		    SelectState := Sel_green;
		    KnobOffset := yPos - (TOP_START + GreenPos);
		elif yPos >= TOP_START + BluePos and
			yPos < TOP_START + KNOB_HEIGHT + BluePos and
			xPos >= BLUE_START and
			xPos < BLUE_START + KNOB_WIDTH then
		    SelectState := Sel_blue;
		    KnobOffset := yPos - (TOP_START + BluePos);
		fi;
	    else
		pos :=
		    if yPos < TOP_START + KnobOffset then
			0
		    elif yPos >= TOP_START + SCALE * 15 + KnobOffset then
			15 * SCALE
		    else
			yPos - KnobOffset - TOP_START
		    fi;
		newValue := 15 - pos / SCALE;
		oldColour :=
		    if IsHalf then
			HalfPtr*.hp_palette[CurrentColour]
		    else
			FullPtr*.fp_palette[CurrentColour]
		    fi;
		case SelectState
		incase Sel_red:
		    if pos ~= RedPos then
			reDrawKnob(RED_START, RedPos, pos);
			RedPos := pos;
		    fi;
		    newColour := oldColour & 0x0ff | (newValue << 8);
		    if newColour ~= oldColour then
			reDrawButton(RED_START, newValue);
		    fi;
		incase Sel_green:
		    if pos ~= GreenPos then
			reDrawKnob(GREEN_START, GreenPos, pos);
			GreenPos := pos;
		    fi;
		    newColour := oldColour & 0xf0f | (newValue << 4);
		    if newColour ~= oldColour then
			reDrawButton(GREEN_START, newValue);
		    fi;
		incase Sel_blue:
		    if pos ~= BluePos then
			reDrawKnob(BLUE_START, BluePos, pos);
			BluePos := pos;
		    fi;
		    newColour := oldColour & 0xff0 | newValue;
		    if newColour ~= oldColour then
			reDrawButton(BLUE_START, newValue);
		    fi;
		esac;
		if IsHalf then
		    HalfPtr*.hp_palette[CurrentColour] := newColour;
		    loadColour(CurrentColour + 16, newColour);
		else
		    FullPtr*.fp_palette[CurrentColour] := newColour;
		    loadColour(CurrentColour, newColour);
		fi;
	    fi;
	elif CycleState = Cy_none then
	    palUnHighLight();
	    case menuSelect(&LastSelect, yPos, xPos)
	    incase 1:
		paletteRead();
	    incase 2:
		paletteReset();
	    incase 3:
		paletteHelp();
		forceChar('\e');
	    incase 4:
		forceChar('q');
	    incase 5:
		startCycle(0);
	    incase 6:
		startCycle(1);
	    incase 7:
		startCycle(2);
	    incase 8:
		startCycle(3);
	    esac;
	fi;
    elif rightDown and CycleState = Cy_none then
	palUnHighLight();
	SelectState := Sel_none;
	if xPos >= 160 - 16 or xPos < DISPLAYCOLS * PATTERNCOLS then
	    LastSelect := 0;
	else
	    case menuSelect(&LastSelect, yPos, xPos)
	    incase 1:
		paletteWrite();
	    esac;
	fi;
    else
	palUnHighLight();
	SelectState := Sel_none;
	LastSelect := 0;
    fi;
corp;

/*
 * paletteCommands - commands during palette editing.
 */

proc paletteCommands()void:
    char CONTROL_C = '\(0x03)';
    uint i;
    char ch;

    installBlinker(palBlink);
    clear();
    if IsHalf then
	LastHalfPalette := CurrentHalf;
	setTopPalette(CurrentHalf);
	setColourCycle(0, 0, 0, 0);
	setColourCycle(1, 0, 0, 0);
	for i from 0 upto 7 do
	    rectangle(i * 8, 160 - 16, i * 8 + 7, 160 - 16 + 7, false, i + 16);
	    rectangle(i * 8, 160 -  8, i * 8 + 7, 160 -  8 + 7, false, i + 24);
	od;
    else
	LastFullPalette := CurrentFull;
	setFullPalette(CurrentFull);
	for i from 0 upto 15 do
	    rectangle(i * 8, 160 - 16, i * 8 + 7, 160 - 16 + 7, false, i);
	    rectangle(i * 8, 160 -  8, i * 8 + 7, 160 -  8 + 7, false, i + 16);
	od;
    fi;
    CurrentColour := 0;
    palMainBox();
    HighLightColour := NOHIGHLIGHT;
    SelectState := Sel_none;
    CycleState := Cy_none;
    LastSelect := 0;
    paletteGadgets();
    installMouser(palMouse);
    pointerOn();
    graphicsDown();
    sayPal();
    while
	ch := getChar();
	if CycleState ~= Cy_none then
	    ch := '\e';
	fi;
	ch ~= 'q' and ch ~= 'Q' and ch ~= CONTROL_C
    do
	case ch
	incase '\e':
	    ;
	incase '\r':
	incase '\n':
	    newLine();
	    sayPal();
	incase '?':
	incase '\(0xfe)':
	    paletteHelp();
	incase 'i':
	incase 'I':
	    paletteRead();
	incase 'o':
	incase 'O':
	    paletteWrite();
	incase 'r':
	incase 'R':
	    paletteReset();
	incase '0':
	    startCycle(0);
	incase '1':
	    startCycle(1);
	incase '2':
	    startCycle(2);
	incase '3':
	    startCycle(3);
	default:
	    palErr("Invalid command");
	esac;
    od;
    newLine();
    pointerOff();
    installMouser(nil);
    installBlinker(nil);
corp;

/*
 * paletteCreate - create a new full or half palette.
 */

proc paletteCreate(bool isHalf)void:
    *CYCLETYPE cy;
    uint i;

    if isHalf then
	if HalfPaletteNext = HALFPALETTEMAX then
	    err("Too many half palettes");
	else
	    CurrentHalf := HalfPaletteNext;
	    HalfPaletteNext := HalfPaletteNext + 1;
	    putString("You are creating half palette \#");
	    putInt(CurrentHalf);
	    putString(".\n");
	    HalfPtr := &HalfPalette[CurrentHalf];
	    HalfPtr*.hp_palette := DEFAULT_HALF_PALETTE;
	    cy := &HalfPtr*.hp_cycles[0];
	    for i from 0 upto 1 do
		cy*.cy_firstColour := 0;
		cy*.cy_lastColour := 0;
		cy*.cy_period := 0;
		cy := cy + sizeof(CYCLETYPE);
	    od;
	    IsHalf := true;
	    paletteCommands();
	fi;
    else
	if FullPaletteNext = FULLPALETTEMAX then
	    err("Too many full palettes");
	else
	    CurrentFull := FullPaletteNext;
	    FullPaletteNext := FullPaletteNext + 1;
	    putString("You are creating full palette \#");
	    putInt(CurrentFull);
	    putString(".\n");
	    FullPtr := &FullPalette[CurrentFull];
	    FullPtr*.fp_palette := DEFAULT_FULL_PALETTE;
	    cy := &FullPtr*.fp_cycles[0];
	    for i from 0 upto 3 do
		cy*.cy_firstColour := 0;
		cy*.cy_lastColour := 0;
		cy*.cy_period := 0;
		cy := cy + sizeof(CYCLETYPE);
	    od;
	    IsHalf := false;
	    paletteCommands();
	fi;
    fi;
corp;

/*
 * paletteEdit - edit an existing full or half palette.
 */

proc paletteEdit(bool isHalf)void:
    int n;

    if isHalf then
	if getNumber(&n, "Edit which half palette? ", 0, HalfPaletteNext - 1,
		     true) then
	    CurrentHalf := n;
	    HalfPtr := &HalfPalette[n];
	    IsHalf := true;
	    paletteCommands();
	fi;
    else
	if getNumber(&n, "Edit which full palette? ", 0, FullPaletteNext - 1,
		     true) then
	    CurrentFull := n;
	    FullPtr := &FullPalette[n];
	    IsHalf := false;
	    paletteCommands();
	fi;
    fi;
corp;

/*
 * halfScn - scan the uses of this half palette. Return 'true' if it's used.
 */

proc halfScn(HALFPALETTE hp)bool:
    ushort i;
    bool used;

    used := scan("Half palette", "maze",
		 pretend(&Maze[0].ma_palette, *ushort),
		 hp, MazeNext, sizeof(MAZETYPE)) ~= 0;
    scan("Half palette", "monster",
	 pretend(&Monster[0].m_palette, *ushort),
	 hp, MonsterNext, sizeof(MONSTERTYPE)) ~= 0 or used
corp;

/*
 * fullScn - scan the uses of this full palette. Return 'true' if it's used.
 */

proc fullScn(FULLPALETTE fp)bool:
    ushort i;

    scan("Full palette", "grid",
	 pretend(&Grid[0].g_palette, *ushort),
	 fp, GridNext, sizeof(GRIDTYPE)) ~= 0
corp;

/*
 * paletteDelete - attempt to delete a full or half palette.
 */

proc paletteDelete(bool isHalf)void:
    int n;

    if isHalf then
	if HalfPaletteNext = 1 then
	    err("Can't delete last half palette");
	elif getNumber(&n, "Delete which half palette? ", 1,
		       HalfPaletteNext - 1, true) then
	    if not halfScn(n) then
		decr(pretend(&Maze[0].ma_palette, *ushort), n,
		     MazeNext, sizeof(MAZETYPE));
		decr(pretend(&Monster[0].m_palette, *ushort), n,
		     MonsterNext, sizeof(MONSTERTYPE));
		HalfPaletteNext := HalfPaletteNext - 1;
		while n ~= HalfPaletteNext do
		    HalfPalette[n] := HalfPalette[n + 1];
		    n := n + 1;
		od;
		putString("Half palette deleted.\n");
	    fi;
	fi;
    else
	if FullPaletteNext = 1 then
	    err("Can't delete last full palette");
	elif getNumber(&n, "Delete which full palette? ", 1,
		       FullPaletteNext - 1, true) then
	    if not fullScn(n) then
		decr(pretend(&Grid[0].g_palette, *ushort), n,
		     GridNext, sizeof(GRIDTYPE));
		FullPaletteNext := FullPaletteNext - 1;
		while n ~= FullPaletteNext do
		    FullPalette[n] := FullPalette[n + 1];
		    n := n + 1;
		od;
		putString("Full palette deleted.\n");
	    fi;
	fi;
    fi;
corp;

/*
 * paletteScan - display the uses of a full or half palette.
 */

proc paletteScan(bool isHalf)void:
    int n;

    if isHalf then
	if getNumber(&n, "Scan which half palette? ", 0,
		     HalfPaletteNext - 1, true) then
	    pretend(halfScn(n), void);
	fi;
    else
	if getNumber(&n, "Scan which full palette? ", 0,
		     FullPaletteNext - 1, true) then
	    pretend(fullScn(n), void);
	fi;
    fi;
corp;
