#/include/explore.g
#cre.g

extern
    getChar()char,
    putString(*char message)void,
    loadFullPalette([PATTERNCOLOURS] uint palette)void,
    loadTopPalette([PICTURECOLOURS] uint palette)void,
    setColourCycle(ushort which, first, last, period)void,

    putInt(uint n)void,
    newLine()void,
    fileDestroy()void,
    fileCreate()void,
    fileOpen1()void,
    fileWrite(byte b)void,
    fileName(*char name, typ)void;

/*
 * err - print the given message with '***' around it.
 */

proc err(*char message)void:

    putString("*** ");
    putString(message);
    putString(". ***\n");
corp;

/*
 * abort - print a message and return to the system.
 */

proc abort(*char message)void:
    extern
	Output()ulong,
	Write(ulong fd; *char m; ulong len)void,
	Exit(ulong status)void,
	amigaTerm()void;
    ulong fd;
    *char p;

    amigaTerm();
    fd := Output();
    Write(fd, "*** ", 4);
    p := message;
    while p* ~= '\e' do
	p := p + 1;
    od;
    Write(fd, message, p - message);
    Write(fd, ". ***\n", 6);
    Exit(30);
corp;

/*
 * getBit - pull a bit out of the BitRaw storage.
 */

proc getBit(BITPOS pos)unsigned 1:

    (BitRaw[BitPosStart + pos / 8] >> (make(pos, ushort) % 8)) & 0x01
corp;

/*
 * putBit - stuff a bit into the BitRaw storage.
 */

proc putBit(BITPOS pos; unsigned 1 b)void:

    BitRaw[BitPosStart + pos / 8] :=
	BitRaw[BitPosStart + pos / 8] & ~(0x01 << (make(pos, ushort) % 8)) |
	    ((b & 0x01) << (make(pos, ushort) % 8))
corp;

/*
 * wordDecr - decrement a set of word pointers.
 */

proc wordDecr(*uint p; uint valu, size, count, uSize)void:

    while count ~= 0 do
	count := count - 1;
	if p* > valu then
	    p* := p* - size;
	fi;
	p := p + uSize;
    od;
corp;

/*
 * shuffleBits - remove a section from the BitPos stuff and shuffle down.
 */

proc shuffleBits(BITPOS start; uint count)void:
    BITPOS end, b;
    uint cnt;
    byte temp;

    cnt := count;
    while cnt ~= 0 do
	cnt := cnt - 1;
	b := cnt + start;
	temp := BitRaw[b];
	end := BitPosNext - count + cnt;
	while b ~= end do
	    BitRaw[b] := BitRaw[b + 1];
	    b := b + 1;
	od;
	BitRaw[end] := temp;
    od;
    BitPosNext := BitPosNext - count;
    wordDecr(pretend(&Grid[0].g_start, *uint),
	     start, count, GridNext, sizeof(GRIDTYPE));
    wordDecr(pretend(&Maze[0].ma_start, *uint),
	     start, count, MazeNext, sizeof(MAZETYPE));
corp;

/*
 * gL1 - get a line with prompt, but can get empty input.
 */

proc gL1(*char prompt)bool:
    extern getString(*char buffer; uint length)void;
    *char p, q;

    putString(prompt);
    getString(&Buffer[0], 127);
    p := &Buffer[0];
    while p* = ' ' do
	p := p + 1;
    od;
    if p* = '\e' then
	false
    else
	q := &Buffer[0];
	while
	    q* := p*;
	    q* ~= '\e'
	do
	    p := p + 1;
	    q := q + 1;
	od;
	true
    fi
corp;

/*
 * getLine - get a line from a given prompt.
 */

proc getLine(*char prompt)void:

    while not gL1(prompt) do
    od;
corp;

/*
 * CAP - capitalize the string in 'Buffer'.
 */

proc CAP()void:
    *char p;

    p := &Buffer[0];
    while p* ~= '\e' do
	if p* >= 'a' and p* <= 'z' then
	    p* := p* - 32;
	fi;
	p := p + 1;
    od;
corp;

/*
 * getFileName - set up a file name for a pattern or picture file.
 */

proc getFileName(*char prompt, typ)bool:

    if gL1(prompt) then
	fileName(&Buffer[0], typ);
	true
    else
	false
    fi
corp;

/*
 * getNumber - prompt for an integer from 1 to the given limit.
 */

proc getNumber(*int pVal; *char prompt;
		      int minimum, maximum; bool inCommand)bool:
    int n;
    char ch;
    bool isNeg, missing;

    if inCommand then
	inCommand := Number* ~= '\e';
    fi;
    while
	missing := false;
	if inCommand then
	    inCommand := false;
	else
	    if gL1(prompt) then
		Number := &Buffer[0];
	    else
		missing := true;
	    fi;
	fi;
	if missing then
	    false
	else
	    if Number* = '-' then
		Number := Number + 1;
		isNeg := true;
	    else
		isNeg := false;
	    fi;
	    n := 0;
	    while
		ch := Number*;
		Number := Number + 1;
		ch >= '0' and ch <= '9'
	    do
		n := n * 10 + (ch - '0');
	    od;
	    if isNeg then
		n := - n;
	    fi;
	    if n > maximum then
		putString("*** Number too large (must be <= ");
		putInt(maximum);
		putString("), try again. ***\n");
		true
	    elif n < minimum then
		putString("*** Number too small (must be >= ");
		putInt(minimum);
		putString("), try again. ***\n");
		true
	    else
		false
	    fi
	fi
    do
    od;
    if missing then
	false
    else
	pVal* := n;
	true
    fi
corp;

/*
 * getNum - use getNumber, but don't allow no entry.
 */

proc getNum(*char prompt; int minimum, maximum)int:
    int n;

    while not getNumber(&n, prompt, minimum, maximum, false) do
	err("You must supply a value here");
    od;
    n
corp;

/*
 * repNum - replace a number, or keep old one.
 */

proc repNum(*char prompt; int minimum, maximum, oldVal)int:
    int n;

    if Editing then
	putString(prompt);
	putInt(oldVal);
	newLine();
	if getNumber(&n, prompt, minimum, maximum, false) then
	    n
	else
	    oldVal
	fi
    else
	getNum(prompt, minimum, maximum)
    fi
corp;

/*
 * inUse - print an in use message.
 */

proc inUse(*char n1; ushort x; *char n2; ushort y)void:

    putString(n1);
    putString(" \#");
    putInt(x);
    putString(" is in use by ");
    putString(n2);
    putString(" \#");
    putInt(y);
    putString(".\n");
corp;

/*
 * scan - scan an array for a particular value, return true if found.
 */

proc scan(*char n1, n2; *ushort p; ushort x, cnt, size)ushort:
    ushort i, found;

    found := 0;
    i := 0;
    while i < cnt do
	if p* = x then
	    inUse(n1, x, n2, i);
	    found := 1;
	fi;
	p := p + size;
	i := i + 1;
    od;
    found
corp;

/*
 * decr - decrement references to values greater than a given one.
 */

proc decr(*ushort p; ushort x, cnt, size)void:

    while cnt ~= 0 do
	cnt := cnt - 1;
	if p* > x then
	    p* := p* - 1;
	fi;
	p := p + size;
    od;
corp;

/*
 * cmdWrd - find an unambiguous command from a list of choices.
 *	    Return 0 for no match or ambiguous, else return the
 *	    number of the choice found.
 */

proc cmdWrd(*char command, choices)ushort:
    *char p;
    ushort which, choice, count;

    count := 0;
    which := 0;
    while choices* ~= '\e' do
	p := command;
	which := which + 1;
	while p* = choices* and choices* ~= '\e' do
	    choices := choices + 1;
	    p := p + 1;
	od;
	if p* = '\e' then
	    count := count + 1;
	    choice := which;
	fi;
	while choices* ~= '\e' do
	    choices := choices + 1;
	od;
	choices := choices + 1;
    od;
    if count = 1 then
	choice
    else
	0
    fi
corp;

/*
 * some routines to print out messages about no such-and-such yet.
 */

proc noOb()void:

    err("You don't have any objects in this world yet");
corp;

proc noMon()void:

    err("You don't have any monsters in this world yet");
corp;

proc noTer()void:

    err("You don't have any territories in this world yet");
corp;

proc noPat()void:

    err("You don't have any patterns in this world yet");
corp;

proc noPic()void:

    err("You don't have any pictures in this world yet");
corp;

proc noGrid()void:

    err("You don't have any grids in this world yet");
corp;

proc noMaze()void:

    err("You don't have any mazes in this world yet");
corp;

/*
 * ask - ask a question - want a yes/no answer.
 */

proc ask(*char question)bool:
    char ch;

    while
	putString(question);
	ch := getChar();
	newLine();
	ch ~= 'y' and ch ~= 'Y' and ch ~= 'n' and ch ~= 'N'
    do
	err("Please answer Y or N");
    od;
    ch = 'y' or ch = 'Y'
corp;

/*
 * menuSelect - return an indication of which of a set of menu buttons has
 *	been selected - don't duplicate the previous one if the button hasn't
 *	been released in the meantime.
 */

proc menuSelect(*uint pLastSelect; uint yPos, xPos)uint:
    uint which;

    if xPos > DISPLAYCOLS * PATTERNCOLS and
	    xPos < DISPLAYCOLS * PATTERNCOLS + 15 then
	which := yPos / PATTERNLINES + 1;
	if yPos <= (which - 1) * PATTERNLINES or
		yPos >= (which - 1) * PATTERNLINES + 15 or
		which = pLastSelect* then
	    0
	else
	    pLastSelect* := which;
	    which
	fi
    else
	0
    fi
corp;

/*
 * setFullPalette - do setup to load a new full palette and cycles.
 */

proc setFullPalette(uint which)void:
    *CYCLETYPE cy;
    uint i;

    loadFullPalette(FullPalette[which].fp_palette);
    cy := &FullPalette[which].fp_cycles[0];
    for i from 0 upto 3 do
	if cy*.cy_lastColour ~= cy*.cy_firstColour then
	    setColourCycle(i, cy*.cy_firstColour, cy*.cy_lastColour,
			   cy*.cy_period);
	fi;
	cy := cy + sizeof(CYCLETYPE);
    od;
corp;

/*
 * setTopPalette - do setup to load a new top palette and cycles.
 */

proc setTopPalette(uint which)void:
    *CYCLETYPE cy;
    uint i;

    loadTopPalette(HalfPalette[which].hp_palette);
    cy := &HalfPalette[which].hp_cycles[0];
    for i from 0 upto 1 do
	if cy*.cy_lastColour ~= cy*.cy_firstColour then
	    setColourCycle(i + 2,
			   cy*.cy_firstColour + PICTURECOLOURS,
			   cy*.cy_lastColour + PICTURECOLOURS,
			   cy*.cy_period);
	fi;
	cy := cy + sizeof(CYCLETYPE);
    od;
corp;
