#drinc:libraries/dos.g
#/include/explore.g

extern
    putString(*char message)void,
    newLine()void,
    getChar()char;

FILENAME FileName;
Handle_t FileHandle;

/*
 * fileMess - output a file name and a message.
 */

proc fileMess(*char message)void:

    putString(&FileName[0]);
    putString(": ");
    putString(message);
    newLine();
corp;

/*
 * fileAbort - an error doing file I/O.
 */

proc fileAbort(*char message)void:
    extern
	amigaTerm()void;
    Handle_t fd;
    *char p;

    amigaTerm();
    fd := Output();
    p := &FileName[0];
    while p* ~= '\e' do
	p := p + 1;
    od;
    pretend(Write(fd, pretend(&FileName[0], *byte), p - &FileName[0]), void);
    pretend(Write(fd, pretend(": ", *byte), 2), void);
    p := message;
    while p* ~= '\e' do
	p := p + 1;
    od;
    pretend(Write(fd, pretend(message, *byte), p - message), void);
    pretend(Write(fd, pretend("\n", *byte), 1), void);
    Exit(30);
corp;

/*
 * fileCreate - create a file.
 */

proc fileCreate()void:

    FileHandle := Open(&FileName[0], MODE_NEWFILE);
    if FileHandle = 0 then
	fileAbort("bad create");
    fi;
    Close(FileHandle);
corp;

/*
 * fileDestroy - destroy a file.
 */

proc fileDestroy()void:

    pretend(DeleteFile(&FileName[0]), void);
corp;

/*
 * fileOpen - try to open a file.
 */

proc fileOpen(bool isRead)bool:

    FileHandle := Open(&FileName[0],
		       if isRead then MODE_READONLY else MODE_NEWFILE fi);
    FileHandle ~= 0
corp;

/*
 * fileOpen1 - open a file, aborting if can't.
 */

proc fileOpen1(bool isRead)void:

    if not fileOpen(isRead) then
	fileAbort("bad open");
    fi;
corp;

/*
 * fileOpen2 - delete, create and open a file.
 */

proc fileOpen2()void:

    fileDestroy();
    fileCreate();
    fileOpen1(false);
corp;

/*
 * fileClose - close a file
 */

proc fileClose()void:

    Close(FileHandle);
corp;

/*
 * fileRead - read a byte from a file.
 */

proc fileRead()byte:
    byte b;

    if Read(FileHandle, &b, 1) ~= 1 then
	Close(FileHandle);
	fileAbort("bad read");
    fi;
    b
corp;

/*
 * fileGet - get several bytes from a file.
 */

proc fileGet(*byte buffer; uint length)void:

    if Read(FileHandle, buffer, length) ~= length then
	Close(FileHandle);
	fileAbort("bad read");
    fi;
corp;

/*
 * fileWrite - write a byte to a file.
 */

proc fileWrite(byte b)void:

    if Write(FileHandle, &b, 1) ~= 1 then
	Close(FileHandle);
	fileAbort("bad write");
    fi;
corp;

/*
 * filePut - put several bytes to a file.
 */

proc filePut(*byte buffer; uint length)void:

    if Write(FileHandle, buffer, length) ~= length then
	Close(FileHandle);
	fileAbort("bad write");
    fi;
corp;

/*
 * fileCRLF - write a CRLF to a file.
 */

proc fileCRLF()void:

    fileWrite('\n' - '\e');
corp;

/*
 * fileChars - write a character string to a file.
 */

proc fileChars(*char p)void:
    *char start;

    start := p;
    while p* ~= '\e' do
	p := p + 1;
    od;
    filePut(pretend(start, *byte), p - start);
corp;

/*
 * fileNum - write a short positive number to a file.
 */

proc fileNum(ushort n)void:

    if n >= 10 then
	fileWrite(n / 10 + '0' - '\e');
    fi;
    fileWrite(n % 10 + '0' - '\e');
corp;

/*
 * fileName - set up a file name.
 */

proc fileName(*char name, defExt)void:
    *char p;

    p := &FileName[0];
    while name* ~= '.' and name* ~= '\e' do
	p* := name*;
	p := p + 1;
	name := name + 1;
    od;
    if name* ~= '.' then
	p* := '.';
	p := p + 1;
	name := defExt;
    fi;
    while
	p* := name*;
	name* ~= '\e'
    do
	p := p + 1;
	name := name + 1;
    od;
corp;

/*
 * fileType - insert a file type.
 */

proc fileType(*char typ)void:
    *char p;

    p := &FileName[0];
    while p* ~= '.' and p* ~= '\e' do
	p := p + 1;
    od;
    p* := '.';
    p := p + 1;
    while
	p* := typ*;
	typ* ~= '\e'
    do
	p := p + 1;
	typ := typ + 1;
    od;
corp;

/*
 * fileDef - set up a default file type.
 */

proc fileDef(*char typ)void:
    *char p;

    p := &FileName[0];
    while p* ~= '.' and p* ~= '\e' do
	p := p + 1;
    od;
    if p* = '\e' then
	fileType(typ);
    fi;
corp;

/*
 * fileSave - save a file name for later use.
 */

proc fileSave(*FILENAME fn)void:

    fn* := FileName;
corp;

/*
 * fileRestore - restore a file name.
 */

proc fileRestore(*FILENAME fn)void:

    FileName := fn*;
corp;

/*
 * fileBackup - rename a file to be of the given extent.
 */

proc fileBackup(*char newExt)void:
    FILENAME backUpName;
    *char p;

    backUpName := FileName;
    p := &backUpName[0];
    while p* ~= '.' and p* ~= '\e' do
	p := p + 1;
    od;
    p* := '.';
    p := p + 1;
    while
	p* := newExt*;
	newExt* ~= '\e'
    do
	p := p + 1;
	newExt := newExt + 1;
    od;
    if not Rename(&FileName[0], &backUpName[0]) then
	fileAbort("bad rename");
    fi;
corp;
