#/include/explore.g
#/include/user.g
#ex.g

extern
    getBit(BITPOS pos)unsigned 1,
    utCmd()void,
    mGen([OCCUPANTCOUNT] OCCUPANTTYPE occupants)void,
    mHit()void,
    stuffCheck()void,
    setFullPalette(FULLPALETTE which)void,
    toggleGraphics()void,

    clear()void,
    installBlinker(proc()void blinker)void,
    scroll(ushort l1, c1, l2, c2; short deltaY, deltaX)void,
    copyIn(ushort l, c, h, w; *uint p)void,
    copyInMasked(ushort l, c, h, w; *uint p, m)void,
    graphicsDown()void,
    beep()void;

*GRIDTYPE GridPtr;
bool Edges, MonsterVisible;

/*
 * getTerritory - pull a terrain value out of the BitRaw data.
 */

proc getTerritory(GRIDLINE l; GRIDCOL c)TERRITORY:
    BITPOS pos;
    uint i;
    TERRITORY t;

    pos := (make(l, BITPOS) * V*.gridCols + c) * TERRAINBITS;
    t := 0;
    for i from 0 upto TERRAINBITS - 1 do
	t := t << 1 | getBit(pos + i);
    od;
    t
corp;

/*
 * lookUpGridSpecial - see if there is a special entry for the current GRID
 *		       at the given position.
 */

proc lookUpGridSpecial(GRIDLINE l; GRIDCOL c)*GRIDSPECIALTYPE:
    *GRIDSPECIALTYPE sp;

    sp := &GridSpecial[0];
    while sp*.gsp_line ~= l or sp*.gsp_col ~= c or
	     sp*.gsp_grid ~= V*.grid do
	sp := sp + sizeof(GRIDSPECIALTYPE);
    od;
    sp
corp;

/*
 * findPattern - find the pattern to display for the given position in
 *		 the current grid.
 */

proc findPattern(GRIDLINE l; GRIDCOL c)PATTERN:
    TERRAIN t;

    t := getTerritory(l, c);
    Territory[if t = T_SPECIAL then
		  lookUpGridSpecial(l, c)*.gsp_territory
	      else
		  GridPtr*.g_territorys[t]
	      fi].t_pattern
corp;

/*
 * insertPattern - do full mode of displaying a pattern on the screen.
 *	This is includes all relevent rounding with neighbours. Passed in
 *	are the 8 neighbouring patterns.
 */

proc insertPattern(ushort row, col;
		   PATTERN tl, t, tr, l, me, r, bl, b, br)void:
    *PATTERNTYPE ptl, pt, ptr, pl, pme, pr, pbl, pb, pbr;

    ptl := &Pattern*[tl];
    pt := &Pattern*[t];
    ptr := &Pattern*[tr];
    pl := &Pattern*[l];
    pme := &Pattern*[me];
    pr := &Pattern*[r];
    pbl := &Pattern*[bl];
    pb := &Pattern*[b];
    pbr := &Pattern*[br];
    /* copy in the whole of the main pattern */
    copyIn(row, col, PATTERNLINES, PATTERNCOLS, &pme*.pa_data[0, 0, 0]);
    /* try to round my top edge */
    if t ~= me and
	    pme*.pa_topRound and t ~= NOPATTERN and pt*.pa_bottomRound and
	    pt*.pa_topRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &pt*.pa_data[0, 0, 0], &Pattern*[0].pa_data[0, 0, 0]);
	if Edges and row ~= 0 and pme*.pa_bottomRound then
	    copyInMasked(row - PATTERNLINES, col, PATTERNLINES, PATTERNCOLS,
		    &pme*.pa_data[0, 0, 0], &Pattern*[2].pa_data[0, 0, 0]);
	fi;
    fi;
    /* try to round my right edge */
    if r ~= me and
	    pme*.pa_rightRound and r ~= NOPATTERN and pr*.pa_leftRound and
	    pr*.pa_rightRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &pr*.pa_data[0, 0, 0], &Pattern*[1].pa_data[0, 0, 0]);
	if Edges and col ~= (DISPLAYCOLS - 1) * PATTERNCOLS and
		pme*.pa_leftRound then
	    copyInMasked(row, col + PATTERNCOLS, PATTERNLINES, PATTERNCOLS,
		     &pme*.pa_data[0, 0, 0], &Pattern*[3].pa_data[0, 0, 0]);
	fi;
    fi;
    /* try to round my bottom edge */
    if b ~= me and
	    pme*.pa_bottomRound and b ~= NOPATTERN and pb*.pa_topRound and
	    pb*.pa_bottomRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &pb*.pa_data[0, 0, 0], &Pattern*[2].pa_data[0, 0, 0]);
	if Edges and row ~= (DISPLAYLINES - 1) * PATTERNLINES and
		pme*.pa_topRound then
	    copyInMasked(row + PATTERNLINES, col, PATTERNLINES, PATTERNCOLS,
		     &pme*.pa_data[0, 0, 0], &Pattern*[0].pa_data[0, 0, 0]);
	fi;
    fi;
    /* try to round my left edge */
    if l ~= me and
	    pme*.pa_leftRound and l ~= NOPATTERN and pl*.pa_rightRound and
	    pl*.pa_leftRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &pl*.pa_data[0, 0, 0], &Pattern*[3].pa_data[0, 0, 0]);
	if Edges and col ~= 0 and pme*.pa_rightRound then
	    copyInMasked(row, col - PATTERNCOLS, PATTERNLINES, PATTERNCOLS,
		     &pme*.pa_data[0, 0, 0], &Pattern*[1].pa_data[0, 0, 0]);
	fi;
    fi;
    /* try to round my top-right corner */
    if tr ~= me and tr ~= NOPATTERN and
	    pt*.pa_group = pr*.pa_group and ptr*.pa_group = pt*.pa_group and
	    pme*.pa_topRound and pme*.pa_rightRound and
	    pt*.pa_bottomRound and pr*.pa_leftRound and
	    ptr*.pa_topRound and ptr*.pa_rightRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &ptr*.pa_data[0, 0, 0], &Pattern*[4].pa_data[0, 0, 0]);
    fi;
    /* try to round my bottom-right corner */
    if br ~= me and br ~= NOPATTERN and
	    pb*.pa_group = pr*.pa_group and pbr*.pa_group = pb*.pa_group and
	    pme*.pa_bottomRound and pme*.pa_rightRound and
	    pb*.pa_topRound and pr*.pa_leftRound and
	    pbr*.pa_bottomRound and pbr*.pa_rightRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &pbr*.pa_data[0, 0, 0], &Pattern*[5].pa_data[0, 0, 0]);
    fi;
    /* try to round my bottom-left corner */
    if bl ~= me and bl ~= NOPATTERN and
	    pb*.pa_group = pl*.pa_group and pbl*.pa_group = pb*.pa_group and
	    pme*.pa_bottomRound and pme*.pa_leftRound and
	    pb*.pa_topRound and pl*.pa_rightRound and
	    pbl*.pa_bottomRound and pbl*.pa_leftRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &pbl*.pa_data[0, 0, 0], &Pattern*[6].pa_data[0, 0, 0]);
    fi;
    /* try to round my top-left corner */
    if tl ~= me and tl ~= NOPATTERN and
	    pt*.pa_group = pl*.pa_group and ptl*.pa_group = pt*.pa_group and
	    pme*.pa_topRound and pme*.pa_leftRound and
	    pt*.pa_bottomRound and pl*.pa_rightRound and
	    ptl*.pa_topRound and ptl*.pa_leftRound then
	copyInMasked(row, col, PATTERNLINES, PATTERNCOLS,
		     &ptl*.pa_data[0, 0, 0], &Pattern*[7].pa_data[0, 0, 0]);
    fi;
corp;

/*
 * displayAt - display the pattern that should be at the given screen
 *	       position in the current grid.
 */

proc displayAt(DISPLAYLINE l1; DISPLAYCOL c1)void:
    SCREENLINE l;
    SCREENCOL c;

    l := l1 + V*.windowLineTop;
    c := c1 + V*.windowColLeft;
    insertPattern(l1 * PATTERNLINES, c1 * PATTERNCOLS,
	    if l = 0 or c = 0 then
		NOPATTERN
	    else
		findPattern(l - 1, c - 1)
	    fi,
	    if l = 0 then
		NOPATTERN
	    else
		findPattern(l - 1, c)
	    fi,
	    if l = 0 or c = V*.gridCols - 1 then
		NOPATTERN
	    else
		findPattern(l - 1, c + 1)
	    fi,
	    if c = 0 then
		NOPATTERN
	    else
		findPattern(l, c - 1)
	    fi,
	    findPattern(l, c),
	    if c = V*.gridCols - 1 then
		NOPATTERN
	    else
		findPattern(l, c + 1)
	    fi,
	    if l = V*.gridLines - 1 or c = 0 then
		NOPATTERN
	    else
		findPattern(l + 1, c - 1)
	    fi,
	    if l = V*.gridLines - 1 then
		NOPATTERN
	    else
		findPattern(l + 1, c)
	    fi,
	    if l = V*.gridLines - 1 or c = V*.gridCols - 1 then
		NOPATTERN
	    else
		findPattern(l + 1, c + 1)
	    fi
    );
corp;

/*
 * gridDisplay - display the current screenful of patterns.
 */

proc gridDisplay()void:
    DISPLAYLINE l;
    DISPLAYCOL c;

    clear();
    for l from 0 upto DISPLAYLINES - 1 do
	for c from 0 upto DISPLAYCOLS - 1 do
	    Edges := l = 0 or c = 0 or
		    l = DISPLAYLINES - 1 or c = DISPLAYCOLS - 1;
	    displayAt(l, c);
	od;
    od;
corp;

/*
 * scrollGrids - scroll the grid region by the given amounts.
 */

proc scrollGrids(short deltaY, deltaX)void:

    scroll(0, 0, DISPLAYLINES * PATTERNLINES - 1,
	   DISPLAYCOLS * PATTERNCOLS - 1, deltaY, deltaX);
corp;

/*
 * cursorUp - move the cursor up, possibly windowing in the grid.
 */

proc cursorUp()void:
    DISPLAYCOL c;

    if V*.gridLine <= V*.windowLineTop + 1 then
	if V*.windowLineTop ~= 0 then
	    V*.windowLineTop := V*.windowLineTop - 1;
	    scrollGrids(-PATTERNLINES, 0);
	    for c from 0 upto DISPLAYCOLS - 1 do
		displayAt(0, c);
	    od;
	fi;
    fi;
corp;

/*
 * cursorDown - move the cursor down, possibly windowing in the grid.
 */

proc cursorDown()void:
    DISPLAYCOL c;

    if V*.gridLine >= V*.windowLineTop + (DISPLAYLINES - 2) then
	if V*.windowLineTop ~= V*.gridLines - DISPLAYLINES then
	    V*.windowLineTop := V*.windowLineTop + 1;
	    scrollGrids(PATTERNLINES, 0);
	    for c from 0 upto DISPLAYCOLS - 1 do
		displayAt(DISPLAYLINES - 1, c);
	    od;
	fi;
    fi;
corp;

/*
 * cursorLeft - move the cursor left, possibly windowing in the grid.
 */

proc cursorLeft()void:
    DISPLAYLINE l;

    if V*.gridCol <= V*.windowColLeft + 1 then
	if V*.windowColLeft ~= 0 then
	    V*.windowColLeft := V*.windowColLeft - 1;
	    scrollGrids(0, -PATTERNCOLS);
	    for l from 0 upto DISPLAYLINES - 1 do
		displayAt(l, 0);
	    od;
	fi;
    fi;
corp;

/*
 * cursorRight - move the cursor right, possibly windowing in the grid.
 */

proc cursorRight()void:
    DISPLAYLINE l;

    if V*.gridCol >= V*.windowColLeft + (DISPLAYCOLS - 2) then
	if V*.windowColLeft ~= V*.gridCols - DISPLAYCOLS then
	    V*.windowColLeft := V*.windowColLeft + 1;
	    scrollGrids(0, PATTERNCOLS);
	    for l from 0 upto DISPLAYLINES - 1 do
		displayAt(l, DISPLAYCOLS - 1);
	    od;
	fi;
    fi;
corp;

/*
 * gridCursor - display the cursor during grid play.
 */

proc gridCursor(PATTERN p; [PATTERNMLINES] uint mask)void:

    copyInMasked(
	(V*.gridLine - V*.windowLineTop) * PATTERNLINES,
	(V*.gridCol - V*.windowColLeft) * PATTERNCOLS,
	PATTERNLINES, PATTERNCOLS,
	&Pattern*[p].pa_data[0, 0, 0], &mask[0]);
corp;

/*
 * gridStandardCursor - display the standard grid cursor.
 */

proc gridStandardCursor()void:

    gridCursor(V*.cursorPattern, CursorMask*);
corp;

/*
 * gridMonsterCursor - display the monster as a cursor.
 */

proc gridMonsterCursor()void:

    gridCursor(V*.monsterData.m_pattern, MonsterMask*);
corp;

/*
 * gridUnCursor - remove the grid cursor.
 */

proc gridUnCursor()void:

    displayAt(V*.gridLine - V*.windowLineTop,
	      V*.gridCol - V*.windowColLeft);
corp;

/*
 * gridTryMove - try grid moving with the given increments.
 */

proc gridTryMove(short lineInc, colInc; proc()void cursorMove)void:
    GRIDLINE l;
    GRIDCOL c;
    TERRAIN t;
    TERRITORY ter;
    ushort group;

    if (V*.gridLine ~= 0 or lineInc ~= -1) and
	    (V*.gridCol ~= 0 or colInc ~= -1) then
	l := V*.gridLine + lineInc;
	c := V*.gridCol + colInc;
	if c < V*.gridCols and l < V*.gridLines then
	    mHit();
	    cursorMove();
	    t := getTerritory(l, c);
	    if
		if t = T_SPECIAL then
		    ter := lookUpGridSpecial(l, c)*.gsp_territory;
		    group := Pattern*[Territory[ter].t_pattern].pa_group;
		    V*.GridSpec = nil or V*.GridSpec(l, c, ter, group)
		else
		    ter := GridPtr*.g_territorys[t];
		    group := Pattern*[Territory[ter].t_pattern].pa_group;
		    V*.GridNorm = nil or V*.GridNorm(l, c, ter, group)
		fi
	    then
		gridUnCursor();
		V*.gridLine := l;
		V*.gridCol := c;
		V*.gridTerritory := ter;
		if lineInc ~= 0 or colInc ~= 0 then
		    V*.underAttack := false;
		fi;
		gridStandardCursor();
		mGen(Territory[ter].t_occupants);
		stuffCheck();
		if V*.GridMoved ~= nil then
		    V*.GridMoved();
		fi;
	    fi;
	else
	    beep();
	fi;
    else
	beep();
    fi;
corp;

/*
 * gridBlink - called to blink any cursor during grid play.
 */

proc gridBlink()void:

    if V*.underAttack then
	MonsterVisible := not MonsterVisible;
	gridUnCursor();
	if MonsterVisible then
	    gridMonsterCursor();
	else
	    gridStandardCursor();
	fi;
    fi;
corp;

/*
 * gridStandard - process a standard command while on grid.
 */

proc gridStandard(char ch)void:

    case ch
    incase '8':
	gridTryMove(-1, 0, cursorUp);
    incase '6':
	gridTryMove(0, +1, cursorRight);
    incase '2':
	gridTryMove(+1, 0, cursorDown);
    incase '4':
	gridTryMove(0, -1, cursorLeft);
    incase '5':
	gridTryMove(0, 0, gridUnCursor);
    incase '.':
	utCmd();
    incase '\r':
    incase '\n':
    incase ' ':
	toggleGraphics();
    default:
	beep();
    esac;
corp;

/*
 * buildCursorMask - build a mask for a grid cursor.
 */

proc buildCursorMask([PATTERNMLINES] uint mask; PATTERN p)void:
    uint i, j, k;

    for i from 0 upto PATTERNMLINES - 1 do
	j := 0;
	for k from 0 upto PATTERNDEPTH - 1 do
	    j := j | Pattern*[p].pa_data[k, i, 0];
	od;
	mask[i] := j;
    od;
corp;

/*
 * setNewCursor - setup for using a new pattern as a cursor.
 */

proc setNewCursor(PATTERN p)void:

    gridUnCursor();
    V*.cursorPattern := p;
    buildCursorMask(CursorMask*, p);
    gridStandardCursor();
corp;

/*
 * setNewMonster - called when a monster is appearing on a grid.
 */

proc setNewMonster()void:

    MonsterVisible := true;
    buildCursorMask(MonsterMask*, V*.monsterData.m_pattern);
    gridUnCursor();
    gridMonsterCursor();
corp;

/*
 * removeMonster - called when the monster has gone away.
 */

proc removeMonster()void:

    if V*.onGrid and MonsterVisible then
	gridUnCursor();
	gridStandardCursor();
    fi;
corp;

/*
 * gridSetup - set up for using the current grid.
 */

proc gridSetup()void:

    V*.onGrid := true;
    V*.underAttack := false;
    GridPtr := &Grid[V*.grid];
    V*.gridLines := GridPtr*.g_lines;
    V*.gridCols := GridPtr*.g_cols;
    BitPosStart := GridPtr*.g_start;
    setFullPalette(GridPtr*.g_palette);
    gridDisplay();
    setNewCursor(V*.cursorPattern);
    gridStandardCursor();
    installBlinker(gridBlink);
    graphicsDown();
corp;
