#/include/explore.g
#/include/user.g
#ex.g

extern
    getBit(BITPOS pos)unsigned 1,
    utCmd()void,
    mGen(*OCCUPANTTYPE oc)void,
    mHit()void,
    stuffCheck()void,
    setBottomPalette(HALFPALETTE hp)void,
    setTopPalette(HALFPALETTE hp)void,
    toggleGraphics()void,

    clear()void,
    line(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    beep()void,
    picParameters(ushort fgColour, bgColour, pattern)void,
    picMove(ushort l, c)void,
    picDraw(ushort l, c)void,
    picEnd()void,
    graphicsDown()void,
    installBlinker(proc()void blinker)void;

*MAZETYPE MazePtr;			/* pointer to current maze */

/* values used in 'mazeDisplay' and its associated routines: */

ushort
    CL = PICTURELINES * 2 / 3,
    CC = PICTURECOLS / 2;

MAZELINE DisplayLine, LeftLine, RightLine;
MAZECOL DisplayCol, LeftCol, RightCol;
byte Config;
short
    LeftLineInc, LeftColInc, RightLineInc, RightColInc,
    HallLineInc, HallColInc,
    WallLineInc, WallColInc;
ushort
    SideShift, WallShift,
    TopThis, BottomThis, SideThis,
    TopNext, BottomNext, SideNext,
    FrontVert, FrontHor,
    SideInnerHor, SideOuterHor,
    TopInnerVert, TopOuterVert, BottomInnerVert, BottomOuterVert,
    TopInnerHor, TopOuterHor,
    TopLeftLine, TopLeftCol, BottomRightLine, BottomRightCol,
    TempLine0,TempLine1,TempLine2,TempLine3,TempLine4,TempLine5,TempLine6,
    TempCol1, TempCol2, TempCol3, TempCol4, TempCol5, TempCol6;
bool LeftBlocked, RightBlocked;

/*
 * mazeBlink - dummy blinker routine - do nothing.
 */

proc mazeBlink()void:

corp;

/*
 * gConf - pull a Configuration value out of the BitRaw data.
 */

proc gConf(MAZELINE l; MAZECOL c)ushort:
    BITPOS pos;
    ushort i, res;

    if l >= V*.mazeLines or c >= V*.mazeCols then
	(M_EMPTY - M_EMPTY) << 4 | (W_WALL - W_WALL) << 2 | (W_WALL - W_WALL)
    else
	pos := (make(l, BITPOS) * V*.mazeCols + c) * MAZEBITS;
	res := 0;
	for i from 0 upto MAZEBITS - 1 do
	    res := getBit(pos) | (res << 1);
	    pos := pos + 1;
	od;
	res
    fi
corp;

/*
 * picDisp - display a picture on the screen.
 */

proc picDisp(PICTURE p)void:
    *POLYGONTYPE pg;
    *VERTEXTYPE v;
    uint i, count;

    pg := &Polygon[Picture[p].p_start];
    for i from 1 upto Picture[p].p_polyCount do
	picParameters(pg*.pg_fgColour + 16, pg*.pg_bgColour + 16,
		      pg*.pg_pattern);
	v := &Vertex[pg*.pg_start];
	picMove(v*.v_line, v*.v_column);
	for count from 2 upto pg*.pg_vertexCount do
	    v := v + sizeof(VERTEXTYPE);
	    picDraw(v*.v_line, v*.v_column);
	od;
	picEnd();
	pg := pg + sizeof(POLYGONTYPE);
    od;
corp;

/*
 * drawLine - interface routine - expensive but it'll do for now.
 */

proc drawLine(MAZELINE l1; MAZECOL c1; MAZELINE l2; MAZECOL c2)void:

    line(l1, c1, l2, c2, false, 12);
corp;

/*
 * ladderLine - same, but when drawing a ladder.
 */

proc ladderLine(MAZELINE l1; MAZECOL c1; MAZELINE l2; MAZECOL c2)void:

    line(l1, c1, l2, c2, false, 2);
corp;

/*
 * ladderDown - draw a ladder going down.
 */

proc ladderDown()void:

    ladderLine(TempLine4, TempCol2, TempLine4, TempCol3);
    ladderLine(TempLine4, TempCol3, TempLine5, TempCol4);
    ladderLine(TempLine5, TempCol4, TempLine5, TempCol1);
    ladderLine(TempLine5, TempCol1, TempLine4, TempCol2);
    TempLine1 := BottomRightLine - (TempLine4 - BottomRightLine + 1) / 2;
    ladderLine(TempLine1, TempCol5, TempLine5, TempCol5);
    ladderLine(TempLine1, TempCol6, TempLine5, TempCol6);
    TempLine1 := BottomRightLine + (TempLine4 - BottomRightLine + 1) / 2;
    ladderLine(TempLine1, TempCol5, TempLine1, TempCol6);
corp;

/*
 * ladderUp - draw a ladder going up.
 */

proc ladderUp()void:

    TempLine1 := CL - TopOuterVert;
    TempLine2 := CL - TopInnerVert;
    ladderLine(TempLine1, TempCol1, TempLine1, TempCol4);
    ladderLine(TempLine1, TempCol4, TempLine2, TempCol3);
    ladderLine(TempLine2, TempCol3, TempLine2, TempCol2);
    ladderLine(TempLine2, TempCol2, TempLine1, TempCol1);
    ladderLine(TempLine1, TempCol5, TempLine4, TempCol5);
    ladderLine(TempLine1, TempCol6, TempLine4, TempCol6);
    TempLine5 := (TempLine4 - TempLine1 + 3) / 6;
    if TempLine5 ~= 0 then
	while
	    TempLine4 := TempLine4 - TempLine5;
	    TempLine4 > TempLine1
	do
	    ladderLine(TempLine4, TempCol5, TempLine4, TempCol6);
	od;
    fi;
corp;

/*
 * mazeDisplayInit - initialize for mazeDisplay.
 */

proc mazeDisplayInit()void:
    short
	LeftHallLineInc, LeftHallColInc,
	RightHallLineInc, RightHallColInc;

    clear();
    TopThis := PICTURELINES * 2 / 3;
    BottomThis := PICTURELINES / 3;
    SideThis := PICTURECOLS / 2;
    FrontVert := PICTURELINES / 6;
    FrontHor := PICTURECOLS / 12;
    SideInnerHor := PICTURECOLS * 315 / 1000;
    SideOuterHor := PICTURECOLS * 396 / 1000;
    TopInnerVert := PICTURELINES * 42 / 100;
    TopOuterVert := PICTURELINES * 53 / 100;
    BottomInnerVert := PICTURELINES * 209 / 1000;
    BottomOuterVert := PICTURELINES * 265 / 1000;
    TopInnerHor := PICTURECOLS * 105 / 1000;
    TopOuterHor := PICTURECOLS * 131 / 1000;
    case V*.heading
    incase NORTH:
	HallLineInc := -1;
	LeftHallColInc := -1;
	SideShift := 0;
	LeftLineInc := 0;
	LeftColInc := 0;
	RightLineInc := 0;
	RightHallLineInc := 0;
	HallColInc := 0;
	WallLineInc := 0;
	WallColInc := 0;
	LeftHallLineInc := 0;
	RightColInc := +1;
	RightHallColInc := +1;
	WallShift := 2;
    incase EAST:
	LeftHallLineInc := -1;
	WallShift := 0;
	LeftLineInc := 0;
	LeftColInc := 0;
	RightColInc := 0;
	HallLineInc := 0;
	WallLineInc := 0;
	LeftHallColInc := 0;
	RightHallColInc := 0;
	HallColInc := +1;
	WallColInc := +1;
	RightHallLineInc := +1;
	RightLineInc := +1;
	SideShift := 2;
    incase SOUTH:
	RightHallColInc := -1;
	SideShift := 0;
	LeftLineInc := 0;
	RightLineInc := 0;
	RightColInc := 0;
	HallColInc := 0;
	WallColInc := 0;
	LeftHallLineInc := 0;
	RightHallLineInc := 0;
	LeftColInc := +1;
	HallLineInc := +1;
	WallLineInc := +1;
	LeftHallColInc := +1;
	WallShift := 2;
    incase WEST:
	HallColInc := -1;
	RightHallLineInc := -1;
	WallShift := 0;
	LeftColInc := 0;
	RightLineInc := 0;
	RightColInc := 0;
	HallLineInc := 0;
	WallLineInc := 0;
	WallColInc := 0;
	LeftHallColInc := 0;
	RightHallColInc := 0;
	LeftLineInc := +1;
	LeftHallLineInc := +1;
	SideShift := 2;
    esac;
    DisplayLine := V*.mazeLine;
    LeftLine := DisplayLine + LeftHallLineInc;
    RightLine := DisplayLine + RightHallLineInc;
    DisplayCol := V*.mazeCol;
    LeftCol := DisplayCol + LeftHallColInc;
    RightCol := DisplayCol + RightHallColInc;
    LeftBlocked := true;
    RightBlocked := true;
corp;

/*
 * mazeDisplayLeft - display walls/doors off to the left of main view.
 */

proc mazeDisplayLeft()void:

    TempCol1 := CC - SideThis;
    Config := (gConf(DisplayLine + LeftLineInc, DisplayCol + LeftColInc) >>
	       SideShift) & 0x03;
    if Config ~= W_EMPTY - W_WALL then
	drawLine(TempLine1, TempCol1, TopLeftLine, TopLeftCol);
	drawLine(TopLeftLine, TopLeftCol, BottomRightLine, TopLeftCol);
	drawLine(BottomRightLine, TopLeftCol, TempLine6, TempCol1);
	drawLine(TempLine6, TempCol1, TempLine1, TempCol1);
	if Config ~= W_WALL - W_WALL then
	    TempCol2 := CC - SideOuterHor;
	    TempCol3 := CC - SideInnerHor;
	    drawLine(TempLine5, TempCol2, TempLine2, TempCol2);
	    drawLine(TempLine2, TempCol2, TempLine3, TempCol3);
	    drawLine(TempLine3, TempCol3, TempLine4, TempCol3);
	fi;
	LeftBlocked := true;
    else
	Config := (gConf(LeftLine + WallLineInc, LeftCol + WallColInc) >>
		   WallShift) & 0x03;
	LeftBlocked :=
	    if Config ~= W_EMPTY - W_WALL then
		drawLine(TopLeftLine, TopLeftCol,
			 BottomRightLine, TopLeftCol);
		TempCol2 := TopLeftCol - SideNext - SideNext + 1;
		if LeftBlocked then
		    drawLine(TopLeftLine, TopLeftCol, TopLeftLine, TempCol1);
		    drawLine(BottomRightLine, TopLeftCol,
			     BottomRightLine, TempCol1);
		else
		    drawLine(TopLeftLine, TopLeftCol, TopLeftLine, TempCol2);
		    drawLine(TopLeftLine, TempCol2,
			     BottomRightLine, TempCol2);
		    drawLine(BottomRightLine, TempCol2,
			     BottomRightLine, TopLeftCol);
		fi;
		if Config ~= W_WALL - W_WALL then
		    TempCol3 := TopLeftCol - SideNext + FrontHor;
		    drawLine(BottomRightLine, TempCol3, TempLine0, TempCol3);
		    TempCol4 :=
			if LeftBlocked then
			    TempCol1
			else
			    TempCol2 := TopLeftCol - SideNext - FrontHor + 1;
			    drawLine(BottomRightLine, TempCol2,
				     TempLine0, TempCol2);
			    TempCol2
			fi;
		    drawLine(TempLine0, TempCol3, TempLine0, TempCol4);
		fi;
		true
	    else
		false
	    fi;
    fi;
corp;

/*
 * mazeDisplayRight - display walls/doors off to the right of main view.
 */

proc mazeDisplayRight()void:

    TempCol1 := CC - 1 + SideThis;
    Config := (gConf(DisplayLine + RightLineInc, DisplayCol + RightColInc) >>
	       SideShift) & 0x03;
    if Config ~= W_EMPTY - W_WALL then
	drawLine(TempLine1, TempCol1, TopLeftLine, BottomRightCol);
	drawLine(TopLeftLine, BottomRightCol,
		 BottomRightLine, BottomRightCol);
	drawLine(BottomRightLine, BottomRightCol, TempLine6, TempCol1);
	drawLine(TempLine6, TempCol1, TempLine1, TempCol1);
	if Config ~= W_WALL - W_WALL then
	    TempCol2 := CC - 1 + SideOuterHor;
	    TempCol3 := CC - 1 + SideInnerHor;
	    drawLine(TempLine5, TempCol2, TempLine2, TempCol2);
	    drawLine(TempLine2, TempCol2, TempLine3, TempCol3);
	    drawLine(TempLine3, TempCol3, TempLine4, TempCol3);
	fi;
	RightBlocked := true;
    else
	Config := (gConf(RightLine + WallLineInc, RightCol + WallColInc) >>
		   WallShift) & 0x03;
	RightBlocked :=
	    if Config ~= W_EMPTY - W_WALL then
		drawLine(TopLeftLine, BottomRightCol,
			 BottomRightLine, BottomRightCol);
		TempCol2 := BottomRightCol + SideNext + SideNext - 1;
		if RightBlocked then
		    drawLine(TopLeftLine, BottomRightCol,
			     TopLeftLine, TempCol1);
		    drawLine(BottomRightLine, BottomRightCol,
			     BottomRightLine, TempCol1);
		else
		    drawLine(TopLeftLine, BottomRightCol,
			     TopLeftLine, TempCol2);
		    drawLine(TopLeftLine, TempCol2,
			     BottomRightLine, TempCol2);
		    drawLine(BottomRightLine, TempCol2,
			     BottomRightLine, BottomRightCol);
		fi;
		if Config ~= W_WALL - W_WALL then
		    TempCol3 := BottomRightCol + SideNext - FrontHor;
		    drawLine(BottomRightLine, TempCol3, TempLine0, TempCol3);
		    TempCol4 :=
			if RightBlocked then
			    TempCol1
			else
			    TempCol2 := BottomRightCol+SideNext+FrontHor - 1;
			    drawLine(BottomRightLine, TempCol2,
				     TempLine0, TempCol2);
			    TempCol2
			fi;
		    drawLine(TempLine0, TempCol3, TempLine0, TempCol4);
		fi;
		true
	    else
		false
	    fi;
    fi;
corp;

/*
 * mazeDisplayStep - cycle one step in maze display.
 */

proc mazeDisplayStep()void:

    DisplayLine := HallLineInc + DisplayLine;
    DisplayCol := HallColInc + DisplayCol;
    LeftLine := HallLineInc + LeftLine;
    LeftCol := HallColInc + LeftCol;
    RightLine := HallLineInc + RightLine;
    RightCol := HallColInc + RightCol;
    TopThis := TopNext;
    BottomThis := BottomNext;
    SideThis := SideNext;
    FrontVert := (FrontVert + 1) / 2;
    FrontHor := (FrontHor + 1) / 2;
    SideInnerHor := (SideInnerHor + 1) / 2;
    SideOuterHor := (SideOuterHor + 1) / 2;
    TopInnerVert := BottomInnerVert;
    TopOuterVert := BottomOuterVert;
    BottomInnerVert := (BottomInnerVert + 1) / 2;
    BottomOuterVert := (BottomOuterVert + 1) / 2;
    TopInnerHor := (TopInnerHor + 1) / 2;
    TopOuterHor := (TopOuterHor + 1) / 2;
corp;

/*
 * mazeDisplay - display the current maze view.
 */

proc mazeDisplay()void:
    ushort i;

    mazeDisplayInit();
    i := 0;
    while i ~= V*.maxView do
	i := i + 1;
	TopNext := (TopThis + 1) / 2;
	BottomNext := (BottomThis + 1) / 2;
	SideNext := (SideThis + 1) / 2;
	TopLeftLine := CL - TopNext;
	TopLeftCol := CC - SideNext;
	BottomRightLine := CL - 1 + BottomNext;
	BottomRightCol := CC - 1 + SideNext;
	TempLine0 := CL - FrontVert;
	Config :=
	    (gConf(DisplayLine + WallLineInc, DisplayCol + WallColInc) >>
	     WallShift) & 0x03;
	if Config ~= W_EMPTY - W_WALL then
	    drawLine(BottomRightLine, TopLeftCol, TopLeftLine, TopLeftCol);
	    drawLine(TopLeftLine, TopLeftCol, TopLeftLine, BottomRightCol);
	    drawLine(TopLeftLine, BottomRightCol,
		     BottomRightLine, BottomRightCol);
	    drawLine(BottomRightLine, BottomRightCol,
		     BottomRightLine, TopLeftCol);
	    if Config ~= W_WALL - W_WALL then
		TempCol1 := CC - FrontHor;
		TempCol2 := CC - 1 + FrontHor;
		drawLine(BottomRightLine, TempCol1, TempLine0, TempCol1);
		drawLine(TempLine0, TempCol1, TempLine0, TempCol2);
		drawLine(TempLine0, TempCol2, BottomRightLine, TempCol2);
	    fi;
	    i := V*.maxView;
	fi;
	TempLine1 := CL - TopThis;
	TempLine2 := CL - BottomOuterVert;
	TempLine3 := CL - BottomInnerVert;
	TempLine4 := CL - 1 + BottomInnerVert;
	TempLine5 := CL - 1 + BottomOuterVert;
	TempLine6 := CL - 1 + BottomThis;
	mazeDisplayLeft();
	mazeDisplayRight();
	TempCol1 := CC - TopOuterHor;
	TempCol2 := CC - TopInnerHor;
	TempCol3 := CC - 1 + TopInnerHor;
	TempCol4 := CC - 1 + TopOuterHor;
	TempCol6 := (TempCol3 - TempCol2) / 4;
	TempCol5 := TempCol2 + TempCol6;
	TempCol6 := TempCol3 - TempCol6;
	Config := gConf(DisplayLine, DisplayCol) & 0x30;
	if Config = (M_LADDER_DOWN - M_EMPTY) << 4 then
	    ladderDown();
	elif Config = (M_LADDER_UP - M_EMPTY) << 4 then
	    ladderUp();
	fi;
	mazeDisplayStep();
    od;
    if V*.underAttack then
	setTopPalette(V*.monsterData.m_palette);
	picDisp(V*.monsterData.m_picture);
    fi;
corp;

/*
 * mazeMonsCheck - local routine to avoid duplication.
 */

proc mmGen()void:

    mGen(&MazePtr*.ma_occupants[0]);
corp;

proc mazeMonsCheck()void:

    mHit();
    mmGen();
    mazeDisplay();
    if V*.MazeMoved ~= nil then
	V*.MazeMoved();
    fi;
corp;

proc mazeMonsCheck2()void:

    if not V*.onGrid then
	mHit();
	mmGen();
	if V*.underAttack then
	    setTopPalette(V*.monsterData.m_palette);
	    picDisp(V*.monsterData.m_picture);
	fi;
    fi;
corp;

/*
 * mazeForward - try to move forward during maze play.
 */

proc mazeForward()void:
    MAZELINE l, cl;
    MAZECOL c, cc;
    ushort wShift;
    bool oldUA;

    mHit();
    l := V*.mazeLine;
    cl := l;
    c := V*.mazeCol;
    cc := c;
    wShift :=
	case V*.heading
	incase NORTH:
	    l := l - 1;
	    2
	incase EAST:
	    cc := cc + 1;
	    c := c + 1;
	    0
	incase SOUTH:
	    cl := cl + 1;
	    l := l + 1;
	    2
	incase WEST:
	    c := c - 1;
	    0
	esac;
    Config := (gConf(cl, cc) >> wShift) & 0x03;
    if Config = W_WALL - W_WALL then
	beep();
    else
	if (V*.MazeMove = nil or V*.MazeMove(l, c)) and
		(Config ~= W_SPECIAL - W_WALL or
		 V*.MazeSpecDoor = nil or
		 V*.MazeSpecDoor(cl, cc, wShift = 2)) then
	    V*.underAttack := false;
	    V*.mazeCol := c;
	    V*.mazeLine := l;
	    if gConf(l, c) & 0x30 = (M_SPECIAL - M_EMPTY) << 4 then
		if V*.MazeSpec ~= nil then
		    V*.MazeSpec();
		fi;
	    fi;
	    if not V*.onGrid then
		mmGen();
		mazeDisplay();
		oldUA := V*.underAttack;
		stuffCheck();
		if V*.MazeMoved ~= nil then
		    V*.MazeMoved();
		fi;
		if V*.underAttack and not oldUA then
		    setTopPalette(V*.monsterData.m_palette);
		    picDisp(V*.monsterData.m_picture);
		fi;
	    fi;
	fi;
    fi;
corp;

/*
 * mazeStandard - standard commands when playing in a maze.
 */

proc mazeStandard(char ch)void:

    case ch
    incase '8':
	mazeForward();
    incase '4':
	V*.heading :=
	    (V*.heading + range(HEADING) - EAST) % range(HEADING) + NORTH;
	mazeMonsCheck();
    incase '6':
	V*.heading := (V*.heading + 1 - NORTH) % range(HEADING) + NORTH;
	mazeMonsCheck();
    incase '2':
	V*.heading := (V*.heading + 2 - NORTH) % range(HEADING) + NORTH;
	mazeMonsCheck();
    incase '5':
	mazeMonsCheck();
    incase 'u':
    incase '+':
	if gConf(V*.mazeLine, V*.mazeCol) & 0x30 =
		(M_LADDER_UP - M_EMPTY) << 4 then
	    if V*.MazeUp ~= nil then
		V*.MazeUp();
	    fi;
	fi;
	mazeMonsCheck2();
    incase 'd':
    incase '-':
	if gConf(V*.mazeLine, V*.mazeCol) & 0x30 =
		(M_LADDER_DOWN - M_EMPTY) << 4 then
	    if V*.MazeDown ~= nil then
		V*.MazeDown();
	    fi;
	fi;
	mazeMonsCheck2();
    incase '.':
	utCmd();
    incase '\r':
    incase '\n':
    incase ' ':
	toggleGraphics();
    default:
	beep();
    esac;
corp;

/*
 * mazeSetup - set up for using the current maze.
 */

proc mazeSetup()void:

    V*.onGrid := false;
    V*.underAttack := false;
    MazePtr := &Maze[V*.maze];
    V*.mazeLines := MazePtr*.ma_lines;
    V*.mazeCols := MazePtr*.ma_cols;
    BitPosStart := MazePtr*.ma_start;
    setBottomPalette(MazePtr*.ma_palette);
    mazeDisplay();
    installBlinker(mazeBlink);
    graphicsDown();
corp;
