#drinc:exec/miscellaneous.g
#drinc:exec/memory.g
#drinc:libraries/dos.g
#/include/explore.g
#/include/user.g
#ex.g

extern
    setCursorRate(uint rate)void,
    putChar(char ch)void,
    putString(*char st)void,
    newLine()void,
    getChar()char,
    getString(*char buffer; uint length)void,
    printerOn()void,
    printerOff()void,
    installBlinker(proc()void blinker)void,
    loadFullPalette(FULLPALETTEARRAY palette)void,
    loadTopPalette(HALFPALETTEARRAY palette)void,
    loadBottomPalette(HALFPALETTEARRAY palette)void,
    loadColour(uint which, colour)void,
    setColourCycle(ushort which, first, last, period)void,
    line(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    rectangle(ushort l1, c1, l2, c2; bool xor; ushort colour)void,
    pixel(ushort l, c, colour)void,
    beep()void,
    graphicsDown()void,
    toggleGraphics()void,

    gridSetup()void,
    mazeSetup()void,
    gridBlink()void,
    mazeBlink()void,
    setNewMonster()void,
    removeMonster()void,
    mazeDisplay()void,
    picDisp(PICTURE p)void,
    charsLen(*char p)uint,
    cap(char ch)char,
    readOb(*char prompt, noneMess; *OBJECT obp; ushort skip, count)ushort,
    drop(OBJECT o)void,
    pickUp()void,
    hereList()void,
    placeHere(OBJECT o)void;

Handle_t OutputFd, WorldFd, TextFd;
Segment_t WorldSeg;
Vector_t ActualVector;			/* this is the actual one */
*byte DAddr;
uint DLen;
uint LastText;
bool Saved;
[256] char WorldName;

/*
 * random - generate a random 16 bit number in a given range.
 */

proc random(uint rang)uint:

    RandomSeed := RandomSeed * 2113 + 13367;
    RandomSeed := (RandomSeed >> 8) + (RandomSeed << 8);
    RandomSeed % rang
corp;

/*
 * getBit - pull a bit out of the BitRaw storage.
 */

proc getBit(BITPOS pos)unsigned 1:

    (BitRaw[pos / 8 + BitPosStart] >> (make(pos, ushort) % 8)) & 0x01
corp;

/*
 * getName - get the address of a name string.
 */

proc getName(uint pos)*char:

    &Name[pos]
corp;

/*
 * getObject
 */

proc getObject(OBJECT ob)*OBJECTTYPE:

    &Object[ob]
corp;

/*
 * getMonster
 */

proc getMonster(MONSTER mo)*MONSTERTYPE:

    &Monster[mo]
corp;

/*
 * print - print a string with \r\n after it.
 */

proc print(*char mess)void:

    putString(mess);
    newLine();
corp;

/*
 * readChars - read a string with a prompt.
 */

proc readChars(*char prompt, buff; ushort len)void:

    putString(prompt);
    installBlinker(nil);
    getString(buff, len);
    installBlinker(if V*.onGrid then gridBlink else mazeBlink fi);
corp;

/*
 * readNum - read in a number.
 */

proc readNum(*char prompt)uint:
    [80] char lineBuffer;
    *char p;
    uint n;

    while
	readChars(prompt, &lineBuffer[0], 80);
	p := &lineBuffer[0];
	n := 0;
	while p* >= '0' and p* <= '9' do
	    n := (p* - '0') + n * 10;
	    p := p + 1;
	od;
	p* ~= '\e'
    do
	print("Bad number try again.");
    od;
    n
corp;

/*
 * pNumber - print out a number.
 */

proc pNumber(uint n)void:
    [6] char buff;
    *char p;

    p := &buff[5];
    p* := '\e';
    while
	p := p - 1;
	p* := n % 10 + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    putString(p);
corp;

/*
 * rRead - read 128 byte record from given position in the .txt file.
 */

proc rRead(uint record; *byte buffer)bool:

    Seek(TextFd, make(record, ulong) * 128, OFFSET_BEGINNING) ~= -1 and
	Read(TextFd, buffer, 128) = 128
corp;

/*
 * pText1 - print out a text entry from the TXT file.
 */

proc pText1(uint record)void:
    type
	ENTRYTYPE = struct {
	    [2] byte e_record;
	    byte e_offset;
	};
    uint ENTRYCOUNT = 128 / sizeof(ENTRYTYPE);
    [128] char buffer;
    [ENTRYCOUNT] ENTRYTYPE Entry @ buffer;
    uint pos;
    char ch;

    record := record + 1;
    if rRead(record / ENTRYCOUNT, pretend(&Entry[0], *byte)) then
	record := record % ENTRYCOUNT;
	pos := Entry[record].e_offset;
	record := (make(Entry[record].e_record[1], uint) << 8) |
			Entry[record].e_record[0];
	if rRead(record, pretend(&buffer[0], *byte)) then
	    while
		if pos = 128 then
		    pos := 0;
		    if Read(TextFd, pretend(&buffer[0], *byte), 128) = 128 then
			ch := buffer[0];
		    else
			putString("\n*** can't read record of entry ***\n");
			ch := '\e';
		    fi;
		else
		    ch := buffer[pos];
		fi;
		ch ~= '\e'
	    do
		putChar(ch);
		pos := pos + 1;
	    od;
	else
	    putString("\n*** can't read first record of entry ***\n");
	fi;
    else
	putString("\n*** can't read index for entry ***\n");
    fi;
corp;

/*
 * pText - world's entry point into pText1.
 */

proc pText(uint record)void:

    LastText := record;
    pText1(record);
corp;

/*
 * pAttType - print out an attack type.
 */

proc pAttType(ATTACK a)void:

    putString(
	case a
	incase A_PUNCTURE:
	    "puncture"
	incase A_SMASH:
	    "smash"
	incase A_MISSILE:
	    "missile"
	incase A_HEAT:
	    "heat"
	incase A_CHILL:
	    "chill"
	incase A_DISPELL:
	    "dispell"
	esac
    );
corp;

/*
 * sPrt - print statistics.
 */

proc sPrt()void:
    *char p, q;
    uint i, j, len;

    putString("Current statistics:\n\n    Experience: ");
    pNumber(V*.experience);
    putString("\n    Hit Points: ");
    pNumber(V*.health);
    putChar('/');
    pNumber(V*.hitPoints);
    putString("\n    Strength: ");
    pNumber(V*.strength);
    if V*.money ~= 0 then
	putString("\n    Money: ");
	pNumber(V*.money);
    fi;
    if V*.food ~= 0 then
	putString("\n    Food: ");
	pNumber(V*.food);
    fi;
    if V*.carryCount ~= 0 then
	putString("\n    Carrying:");
	if V*.carryWeight ~= 0 then
	    putString(" (total weight ");
	    pNumber(V*.carryWeight);
	    putChar('/');
	    pNumber(V*.maxCarryWeight);
	    putChar(')');
	fi;
	j := 71;
	for i from 0 upto V*.carryCount - 1 do
	    p := getName(getObject(V*.carrying[i])*.ob_name);
	    q := p;
	    while q* ~= '\e' do
		q := q + 1;
	    od;
	    len := q - p;
	    if len + j + 2 > 71 then
		j := 0;
		putString("\n\t");
	    fi;
	    j := len + j + 2;
	    putString(p);
	    if i ~= V*.carryCount - 1 then
		putString(", ");
	    fi;
	od;
    fi;
    if V*.armed then
	putString("\n    Weapon: ");
	putString(getName(getObject(V*.weaponNumber)*.ob_name));
    fi;
    i := 0;
    while i ~= range(ATTACK) - 1 and V*.protection[i] = 0 do
	i := i + 1;
    od;
    if i ~= range(ATTACK) - 1 then
	putString("\n    Protection:\n\t");
	for i from 0 upto range(ATTACK) - 2 do
	    pAttType(i + A_PUNCTURE);
	    putChar(':');
	    if V*.protection[i] < 100 then
		putChar(' ');
	    fi;
	    pNumber(V*.protection[i]);
	    putChar(' ');
	od;
    fi;
    newLine();
    if V*.Stats ~= nil then
	V*.Stats();
    fi;
corp;

/*
 * save - save world state to a file.
 */

proc save()void:
    uint
	PS = sizeof(proc()void),
	PT = PS * NP,
	FS = sizeof(Vector_t) - PT;
    [80] char lineBuffer;
    *char p;
    Handle_t fd;

    readChars("File to use: ", &lineBuffer[0], 80);
    print(
	if lineBuffer[0] ~= '\e' then
	    p := &lineBuffer[0];
	    while p* ~= '\e' do
		p := p + 1;
	    od;
	    p* := '.';
	    (p + 1)* := 'x';
	    (p + 2)* := 's';
	    (p + 3)* := 'v';
	    (p + 4)* := '\e';
	    pretend(DeleteFile(&lineBuffer[0]), void);
	    fd := Open(&lineBuffer[0], MODE_NEWFILE);
	    if fd ~= 0 then
		if V*.Save = nil or V*.Save() then
		    if Write(fd, pretend(V, *byte) + PT, FS) ~= FS or
			    DLen ~= 0 and Write(fd, DAddr, DLen) ~= DLen then
			Close(fd);
			"Save failed."
		    else
			Close(fd);
			V*.gameOver := true;
			Saved := true;
			"Saved."
		    fi
		else
		    Close(fd);
		    "Can't save now."
		fi
	    else
		"Can't open save file."
	    fi
	else
	    "Save aborted."
	fi);
corp;

/*
 * getCarry - get the number of an object being carried.
 */

proc getCarry(*char prompt)ushort:

    if V*.carryCount = 0 then
	print("You aren't carrying anything.");
	255
    else
	readOb(prompt, "You aren't carrying one.",
	       &V*.carrying[0], sizeof(OBJECT), V*.carryCount)
    fi
corp;

/*
 * utCmd - process a utility command.
 */

proc utCmd()void:
    ushort i;

    case cap(getChar())
    incase 'Q':
	V*.gameOver := true;
    incase 'S':
	save();
    incase '.':
	sPrt();
    incase 'L':
	hereList();
    incase 'G':
	pickUp();
    incase 'D':
	i := getCarry("Object to drop? ");
	if i = 255 then
	    print("Nothing dropped.");
	else
	    if V*.Dropped = nil or V*.Dropped(i) then
		drop(i);
		placeHere(i);
		print("Dropped.");
	    fi;
	fi;
    incase 'U':
	i := getCarry("Object to use? ");
	if i = 255 then
	    print("Nothing used.");
	else
	    if V*.Use ~= nil then
		V*.Use(i);
	    fi;
	fi;
    incase 'W':
	i := getCarry("Weapon to wield? ");
	if i = 255 then
	    V*.armed := false;
	    print("Using bare hands.");
	else
	    V*.armed := true;
	    V*.weaponNumber := i;
	fi;
    incase 'R':
	pText1(LastText);
    incase 'P':
	if V*.printing then
	    V*.printing := false;
	    printerOff();
	    print("Printer echo off.");
	else
	    print("Printer echo on.");
	    printerOn();
	    V*.printing := true;
	fi;
    incase 'C':
	setCursorRate(readNum("New cursor blink rate: "));
    incase '?':
	print(
	    "\nUtility commands are: (preceed each with '.')\n\n"
	    "\tq - quit\n"
	    "\ts - save current state\n"
	    "\t. - display status\n"
	    "\tl - look at objects here\n"
	    "\tg - get an object\n"
	    "\td - drop an object\n"
	    "\tu - use an object\n"
	    "\tw - wield a weapon\n"
	    "\tr - repeat last text\n"
	    "\tp - printer toggle\n"
	    "\tc - set cursor blink rate\n"
	);
    default:
	print("Bad utility command. .? for help.");
    esac;
corp;

/*
 * mHit - the monster hits the player.
 */

proc mHit()void:
    uint damage, prot;

    if V*.underAttack and V*.monsterData.m_type ~= A_HARMLESS then
	prot := V*.protection[V*.monsterData.m_type - 1];
	damage := V*.monsterData.m_strength / 4;
	damage := random(damage * 3) + damage;
	if damage < prot then
	    damage := 0;
	else
	    damage := damage - prot;
	fi;
	putString("The ");
	putString(&Name[V*.monsterData.m_name]);
	if damage = 0 then
	    print("'s attack is deflected.");
	else
	    putString(" hits (");
	    pAttType(V*.monsterData.m_type);
	    putString(") for ");
	    pNumber(damage);
	    print(" points.");
	    V*.health :=
		if damage >= V*.health then
		    V*.gameOver := true;
		    0
		else
		    V*.health - damage
		fi;
	fi;
    fi;
corp;

/*
 * monsterAttack - an actual attack by a monster.
 */

proc monsterAttack()void:

    if V*.underAttack then
	if V*.onGrid then
	    setNewMonster();
	fi;
	putString(&Name[V*.monsterData.m_name]);
	putString(" [");
	pNumber((make(V*.monsterData.m_hitPoints, uint) +
		   V*.monsterData.m_strength) / 50);
	print("]");
	if random(V*.monsterData.m_speed) >=
		random(V*.experience / 25 - V*.carryWeight / 4 + 128) then
	    mHit();
	fi;
    fi;
corp;

/*
 * mAtt - call monsterAttack after setting up the monster data.
 */

proc mAtt()void:

    V*.monsterData := Monster[V*.monsterNumber];
    monsterAttack();
corp;

/*
 * mGen - see if a monster should appear.
 */

proc mGen([OCCUPANTCOUNT] OCCUPANTTYPE oc)void:
    *OCCUPANTTYPE o;
    uint rand;
    ushort i;
    bool gotOne;

    o := &oc[0];
    if not V*.underAttack then
	gotOne := false;
	rand := random(255 * OCCUPANTCOUNT) + 1;
	for i from 0 upto OCCUPANTCOUNT - 1 do
	    if not gotOne and rand <= o*.oc_probability then
		V*.monsterNumber := o*.oc_type;
		V*.monsterData := Monster[o*.oc_type];
		V*.underAttack := true;
		if V*.MonsterCheck ~= nil then
		    V*.MonsterCheck();
		fi;
		gotOne := true;
	    else
		rand := rand - o*.oc_probability;
	    fi;
	    o := o + sizeof(OCCUPANTTYPE);
	od;
	if V*.underAttack then
	    monsterAttack();
	fi;
    fi;
corp;

/*
 * pHit - the player hits the monster.
 */

proc pHit(ATTACK typ; uint stren; *char missMess, hitMess)void:
    uint damage, exp;
    int vuln;

    if missMess = nil then
	damage := stren;
	vuln := 0;
    else
	vuln := if typ ~= A_HARMLESS then
		    V*.monsterData.m_vulnerability[typ - 1]
		else
		    0
		fi;
	damage := stren / 4;
	if damage ~= 0 then
	    damage := random(damage * 3) + damage;
	    damage :=
		if damage > 500 then
		    damage / 10 * make(|vuln, uint) / 5
		else
		    damage * make(|vuln, uint) / 50
		fi;
	fi;
    fi;
    if vuln < 0 then
	putString("The ");
	putString(&Name[V*.monsterData.m_name]);
	print(" is strengthened!");
	V*.monsterData.m_hitPoints :=
	    if make(V*.monsterData.m_hitPoints, uint) + damage < 0xff then
		make(damage, ushort) + V*.monsterData.m_hitPoints
	    else
		0xff
	    fi;
    elif damage = 0 then
	print(missMess);
    else
	putString(hitMess);
	pNumber(damage);
	putString(" points.");
	if damage > V*.monsterData.m_hitPoints then
	    print(" - It's dead!");
	    V*.underAttack := false;
	    if V*.KillMonster ~= nil then
		V*.KillMonster();
	    fi;
	    if V*.onGrid then
		removeMonster();
	    else
		mazeDisplay();
	    fi;
	else
	    V*.monsterData.m_hitPoints := V*.monsterData.m_hitPoints - damage;
	    newLine();
	fi;
    fi;
corp;

/*
 * setFullPalette - do setup to load a new full palette and cycles.
 */

proc setFullPalette(FULLPALETTE which)void:
    *CYCLETYPE cy;
    uint i;

    loadFullPalette(FullPalette[which].fp_palette);
    cy := &FullPalette[which].fp_cycles[0];
    for i from 0 upto 3 do
	if cy*.cy_lastColour ~= cy*.cy_firstColour then
	    setColourCycle(i, cy*.cy_firstColour, cy*.cy_lastColour,
			   cy*.cy_period);
	fi;
	cy := cy + sizeof(CYCLETYPE);
    od;
corp;

/*
 * setTopPalette - do setup to load a new top palette and cycles.
 */

proc setTopPalette(HALFPALETTE which)void:
    *CYCLETYPE cy;
    uint i;

    loadTopPalette(HalfPalette[which].hp_palette);
    cy := &HalfPalette[which].hp_cycles[0];
    for i from 0 upto 1 do
	if cy*.cy_lastColour ~= cy*.cy_firstColour then
	    setColourCycle(i + 2,
			   cy*.cy_firstColour + PICTURECOLOURS,
			   cy*.cy_lastColour + PICTURECOLOURS,
			   cy*.cy_period);
	fi;
	cy := cy + sizeof(CYCLETYPE);
    od;
corp;

/*
 * setBottomPalette - do setup to load a new bottom palette and cycles.
 */

proc setBottomPalette(HALFPALETTE which)void:
    *CYCLETYPE cy;
    uint i;

    loadBottomPalette(HalfPalette[which].hp_palette);
    cy := &HalfPalette[which].hp_cycles[0];
    for i from 0 upto 1 do
	if cy*.cy_lastColour ~= cy*.cy_firstColour then
	    setColourCycle(i + 2, cy*.cy_firstColour,
			   cy*.cy_lastColour, cy*.cy_period);
	fi;
	cy := cy + sizeof(CYCLETYPE);
    od;
corp;

/*
 * worldSetup - a world has been read in - now let it initialize.
 */

proc worldSetup()void:
    extern
	gridDisplay()void,
	gridStandard(char ch)void,
	setNewCursor(PATTERN p)void,
	mazeStandard(char ch)void,
	gConf(MAZELINE l; MAZECOL c)ushort,
	carry(OBJECT o)void,
	place(OBJECT o; bool onGr; ushort gridMaze, line, column)void,
	remove(OBJECT o; bool onGr; ushort gridMaze, line, column)void,
	isHere(OBJECT o)bool,
	isCarrying(OBJECT o)bool,
	stuffCheck()void;
    DateStamp_t ds;
    ATTACK a;

    V := &ActualVector;

    V*.PutChar := putChar;
    V*.PMess := putString;
    V*.Pnl := newLine;
    V*.Print := print;
    V*.PNumber := pNumber;
    V*.PText := pText;
    V*.ReadChar := getChar;
    V*.ReadChars := readChars;
    V*.ReadNum := readNum;
    V*.Beep := beep;

    V*.Random := random;
    V*.CharsLen := charsLen;
    V*.Cap := cap;

    V*.GraphicsDown := graphicsDown;
    V*.ToggleGraphics := toggleGraphics;
    V*.Point := pixel;
    V*.Line := line;
    V*.Rectangle := rectangle;
    V*.PicDisp := picDisp;
    V*.SetFullPalette := setFullPalette;
    V*.SetBottomPalette := setBottomPalette;
    V*.SetTopPalette := setTopPalette;
    V*.SetColour := loadColour;
    V*.DisplayImage := nil;				/* */

    V*.GetMonster := getMonster;
    V*.PAttType := pAttType;
    V*.GetName := getName;
    V*.GetObject := getObject;

    V*.Carry := carry;
    V*.Drop := drop;
    V*.Place := place;
    V*.Remove := remove;
    V*.PlaceHere := placeHere;
    V*.IsHere := isHere;
    V*.IsCarrying := isCarrying;
    V*.StuffCheck := stuffCheck;
    V*.HereList := hereList;

    V*.MAtt := mAtt;
    V*.SpecificMAtt := monsterAttack;
    V*.MHit := mHit;
    V*.PHit := pHit;
    V*.RemoveMonster := removeMonster;

    V*.GridSetup := gridSetup;
    V*.GridDisplay := gridDisplay;
    V*.GridStandard := gridStandard;
    V*.SetNewCursor := setNewCursor;
    V*.SaveBackGround := nil;				/* */
    V*.DisplayMonster := nil;				/* */
    V*.RestoreBackGround := nil;			/* */

    V*.MazeSetup := mazeSetup;
    V*.MazeDisplay := mazeDisplay;
    V*.MazeStandard := mazeStandard;
    V*.GConf := gConf;

    DateStamp(&ds);
    RandomSeed := (ds.ds_Days >< ds.ds_Minute >< ds.ds_Tick) | 1;
    LastText := 0;
    V*.printing := false;
    V*.gameOver := false;
    V*.underAttack := false;
    V*.armed := false;
    V*.maxView := MAXVIEW;
    V*.cursorPattern := 0;
    V*.hitPoints := 0;
    V*.strength := 0;
    V*.experience := 0;
    V*.money := 0;
    V*.food := 0;
    V*.health := 0;
    V*.carryWeight := 0;
    V*.carryCount := 0;
    V*.droppedCount := 0;
    V*.maxCarryWeight := 0xffff;
    V*.maxCarryCount := MAXCARRY;
    for a from A_PUNCTURE upto A_DISPELL do
	V*.protection[a - 1] := 0;
    od;

    V*.Restore := nil;
    V*.Save := nil;
    V*.Run := nil;
    V*.Stats := nil;
    V*.GridNorm := nil;
    V*.GridSpec := nil;
    V*.GridMoved := nil;
    V*.MazeMove := nil;
    V*.MazeUp := nil;
    V*.MazeDown := nil;
    V*.MazeMoved := nil;
    V*.MazeSpec := nil;
    V*.MazeSpecDoor := nil;
    V*.MonsterCheck := nil;
    V*.KillMonster := nil;
    V*.Dropped := nil;
    V*.Use := nil;

    /* deep, dark, black magic - call the dynamically loaded code */

    pretend((WorldSeg + 1) << 2, proc(*Vector_t v; **byte pP; *uint pLen)void)
		(V, &DAddr, &DLen);
corp;

/*
 * fileGet - safely read some bytes from the world file.
 */

proc fileGet(*byte buffer; uint length)void:

    if WorldFd ~= 0 then
	if Read(WorldFd, buffer, length) ~= length then
	    Close(WorldFd);
	    WorldFd := 0;
	fi;
    fi;
corp;

/*
 * worldGet - read in the entire world.
 */

proc worldGet()void:

    fileGet(pretend(&NameNext, *byte), sizeof(uint));
    fileGet(pretend(&Name[0], *byte), sizeof(char) * NameNext);
    fileGet(pretend(&GridNext, *byte), sizeof(GRID));
    fileGet(pretend(&Grid[0], *byte),
	      sizeof(GRIDTYPE) * make(GridNext, uint));
    fileGet(pretend(&GridSpecialNext, *byte), sizeof(GRIDSPECIAL));
    fileGet(pretend(&GridSpecial[0], *byte),
	      sizeof(GRIDSPECIALTYPE) * make(GridSpecialNext, uint));
    fileGet(pretend(&TerritoryNext, *byte), sizeof(TERRITORY));
    fileGet(pretend(&Territory[0], *byte),
	      sizeof(TERRITORYTYPE) * make(TerritoryNext, uint));
    fileGet(pretend(&PatternNext, *byte), sizeof(PATTERN));
    fileGet(pretend(&Pattern*[0], *byte),
	      sizeof(PATTERNTYPE) * make(PatternNext, uint));
    fileGet(pretend(&MazeNext, *byte), sizeof(MAZE));
    fileGet(pretend(&Maze[0], *byte),
	      sizeof(MAZETYPE) * make(MazeNext, uint));
    fileGet(pretend(&PictureNext, *byte), sizeof(PICTURE));
    fileGet(pretend(&Picture[0], *byte),
	      sizeof(PICTURETYPE) * make(PictureNext, uint));
    fileGet(pretend(&PolygonNext, *byte), sizeof(POLYGON));
    fileGet(pretend(&Polygon[0], *byte),
	      sizeof(POLYGONTYPE) * make(PolygonNext, uint));
    fileGet(pretend(&VertexNext, *byte), sizeof(VERTEX));
    fileGet(pretend(&Vertex[0], *byte),
	      sizeof(VERTEXTYPE) * make(VertexNext, uint));
    fileGet(pretend(&ObjectNext, *byte), sizeof(OBJECT));
    fileGet(pretend(&Object[0], *byte),
	      sizeof(OBJECTTYPE) * make(ObjectNext, uint));
    fileGet(pretend(&MonsterNext, *byte), sizeof(MONSTER));
    fileGet(pretend(&Monster[0], *byte),
	      sizeof(MONSTERTYPE) * make(MonsterNext, uint));
    fileGet(pretend(&FullPaletteNext, *byte), sizeof(FULLPALETTE));
    fileGet(pretend(&FullPalette[0], *byte),
	      sizeof(FULLPALETTETYPE) * make(FullPaletteNext, uint));
    fileGet(pretend(&HalfPaletteNext, *byte), sizeof(HALFPALETTE));
    fileGet(pretend(&HalfPalette[0], *byte),
	      sizeof(HALFPALETTETYPE) * make(HalfPaletteNext, uint));
    fileGet(pretend(&BitPosNext, *byte), sizeof(BITPOS));
    fileGet(pretend(&BitRaw[0], *byte), BitPosNext);
corp;

/*
 * restore - restore world from file.
 */

proc restore()void:
    uint
	PS = sizeof(proc()void),
	PT = PS * NP,
	FS = sizeof(Vector_t) - PT;
    bool attack;

    if Read(WorldFd, pretend(V, *byte) + PT, FS) ~= FS or
	    DLen ~= 0 and Read(WorldFd, DAddr, DLen) ~= DLen then
	Close(WorldFd);
	WorldFd := 0;
    else
	attack := V*.underAttack;
	if V*.onGrid then
	    gridSetup();
	    if attack then
		setNewMonster();
	    fi;
	else
	    mazeSetup();
	    if attack then
		setTopPalette(V*.monsterData.m_palette);
		picDisp(V*.monsterData.m_picture);
	    fi;
	fi;
	V*.underAttack := attack;
    fi;
corp;

/*
 * say - write a string to the normal output.
 */

proc say(*char message)void:
    *char p;

    p := message;
    while p* ~= '\e' do
	p := p + 1;
    od;
    pretend(Write(OutputFd, pretend(message, *byte), p - message), void);
corp;

/*
 * main1 - inner part of main
 */

proc main1(*char p, q)void:
    extern
	amigaInit()bool,
	amigaTerm()void;

    worldGet();
    if WorldFd ~= 0 then
	Close(WorldFd);
	(q + 1)* := 't';
	(q + 2)* := 'x';
	(q + 3)* := 't';
	TextFd := Open(&WorldName[0], MODE_READONLY);
	if TextFd ~= 0 then
	    (q + 1)* := 'c';
	    (q + 2)* := 'o';
	    (q + 3)* := 'd';
	    WorldSeg := LoadSeg(&WorldName[0]);
	    if WorldSeg ~= 0 then
		if amigaInit() then
		    Saved := false;
		    worldSetup();
		    while p* = ' ' do
			p := p + 1;
		    od;
		    if V*.Run = nil then
			say("No V*.Run scenario entry point - aborting\n");
		    elif p* = '\n' or p* = '\r' or p* = '\e' then
			/* starting a new game */
			V*.Run(false);
		    else
			/* restoring a previous game */
			q := &WorldName[0];
			while p* ~= '\r' and p* ~= '\n' and p* ~= '\e' and
				p* ~= ' ' do
			    q* := p*;
			    q := q + 1;
			    p := p + 1;
			od;
			q* := '.';
			(q + 1)* := 'x';
			(q + 2)* := 's';
			(q + 3)* := 'v';
			(q + 4)* := '\e';
			WorldFd := Open(&WorldName[0], MODE_READONLY);
			if WorldFd ~= 0 then
			    restore();
			    if WorldFd ~= 0 then
				Close(WorldFd);
				pretend(DeleteFile(&WorldName[0]), void);
				if V*.Restore ~= nil then
				    V*.Restore();
				fi;
				V*.Run(true);
			    else
				say("Bad read from save file - aborting\n");
			    fi;
			else
			    say("Can't open save file - aborting\n");
			fi;
		    fi;
		    if V*.health = 0 then
			print("You are dead!\n\nGame over.\n");
		    fi;
		    if not Saved then
			putString("Type a key:");
			pretend(getChar(), void);
		    fi;
		    amigaTerm();
		else
		    say("Amiga specific initialization failed - aborting\n");
		fi;
		if not UnLoadSeg(WorldSeg) then
		    say("Eeek - UnLoadSeg failed\n");
		fi;
	    else
		say("Can't load code file ");
		say(&WorldName[0]);
		say(" - aborting\n");
	    fi;
	    Close(TextFd);
	else
	    say("Can't open data file ");
	    say(&WorldName[0]);
	    say(" - aborting\n");
	fi;
    else
	say("Bad read from data file ");
	say(&WorldName[0]);
	say(" - aborting\n");
    fi;
corp;

/*
 * main - handle parameters (world names)
 */

proc main()void:
    extern
	GetPars(*ulong pLen; **char pP)void;
    *char p, q;
    ulong len;

    if OpenDosLibrary(0) ~= nil then
	if OpenExecLibrary(0) ~= nil then
	    OutputFd := Output();
	    GetPars(&len, &p);
	    while p* = ' ' do
		p := p + 1;
	    od;
	    if p* ~= '\n' and p* ~= '\r' and p* ~= '\e' then
		q := &WorldName[0];
		while p* ~= '\r' and p* ~= '\n' and p* ~= '\e' and p* ~= ' ' do
		    q* := p*;
		    q := q + 1;
		    p := p + 1;
		od;
		q* := '.';
		(q + 1)* := 'w';
		(q + 2)* := 'l';
		(q + 3)* := 'd';
		(q + 4)* := '\e';
		WorldFd := Open(&WorldName[0], MODE_READONLY);
		if WorldFd ~= 0 then
		    Pattern :=
			pretend(AllocMem(PATTERNMAX * sizeof(PATTERNTYPE),
				MEMF_CHIP), *[PATTERNMAX] PATTERNTYPE);
		    if Pattern ~= nil then
			CursorMask :=
			    pretend(AllocMem(PATTERNMLINES * sizeof(uint),
				    MEMF_CHIP), *[PATTERNMLINES] uint);
			if CursorMask ~= nil then
			    MonsterMask :=
				pretend(AllocMem(PATTERNMLINES * sizeof(uint),
					MEMF_CHIP), *[PATTERNMLINES] uint);
			    if MonsterMask ~= nil then
				main1(p, q);
				FreeMem(pretend(MonsterMask, *byte),
					PATTERNMLINES * sizeof(uint));
			    else
				say("Can't allocate chip memory for monster "
				    "mask - aborting\n");
			    fi;
			    FreeMem(pretend(CursorMask, *byte),
				    PATTERNMLINES * sizeof(uint));
			else
			    say("Can't allocate chip memory for cursor mask - "
				"aborting\n");
			fi;
			FreeMem(pretend(Pattern, *byte),
				PATTERNMAX * sizeof(PATTERNTYPE));
		    else
			say("Can't allocate chip memory for patterns - "
			    "aborting\n");
		    fi;
		else
		    say("Can't open data file ");
		    say(&WorldName[0]);
		    say(" - aborting\n");
		fi;
	    else
		say("Use is: explore worldfile [savefile]\n");
	    fi;
	    CloseExecLibrary();
	else
	    say("Can't open exec.library - aborting\n");
	fi;
	CloseDosLibrary();
    fi;
corp;
