/* a utility value needed for file I/O */

uint
    FILENAMELENGTH = 100;

type
    FILENAME = [FILENAMELENGTH] char;

/* adjustable parameters: (affect only storage requirements of world) */

uint
    NAMEMAX = 2500,		/* size of name buffer */
    OCCUPANTCOUNT = 6,		/* number of different occupants/territory */

    PATTERNMAX = 88,		/* max # of patterns in world */
    GRIDSPECIALMAX = 250,	/* max # of special patterns in world */
    GRIDMAX = 50,		/* max # of grids in world */
    PICTUREMAX = 100,		/* max # of pictures in world */
    POLYGONMAX = 1000,		/* max # of picture polygons in world */
    VERTEXMAX = 10000,		/* max # of picture vertexes in world */
    MAZEMAX = 50,		/* max # of mazes in world */
    BITPOSMAX = 10000,		/* max size of all grids/mazes in world */
    OBJECTMAX = 100,		/* max # of object types in world */
    MONSTERMAX = 250,		/* max # of monster types in world */
    TERRITORYMAX = 250, 	/* max # of terrain types in world */
    FULLPALETTEMAX = 10,	/* max # of full palettes in world */
    HALFPALETTEMAX = 50;	/* max # of half palettes in world */

/* specifics of how the program uses the hardware: */

uint
    SCREENLINES = 128,		/* # lines on graphics screen */
    SCREENCOLS = 160,		/* # cols on graphics screen */

    DISPLAYLINES = 8,		/* # of lines of patterns on screen */
    DISPLAYCOLS = 8,		/* # of cols of patterns on screen */

    TERRAINBITS = 5,		/* # bits in grid to indicate pattern */
    MAZEBITS = 6,		/* # bits for maze at each position */

    PATTERNLINES = 16,
    PATTERNCOLS = 16,
    PATTERNMLINES = PATTERNLINES,
    PATTERNMCOLS = PATTERNCOLS / 16,	/* it's an array of words */
    PATTERNDEPTH = 5,
    PATTERNCOLOURS = 1 << PATTERNDEPTH,

    PICTUREDEPTH = 4,
    PICTURECOLOURS = 1 << PICTUREDEPTH,
    PICTURELINES = 128, 	/* # lines in picture area */
    PICTURECOLS = 128,		/* # cols in picture area */

    MAXVIEW = 5;		/* max # maze spaces visible on screen */

type
    ATTACK = enum {
	A_HARMLESS,
	A_PUNCTURE,
	A_SMASH,
	A_MISSILE,
	A_HEAT,
	A_CHILL,
	A_DISPELL
    },

    MAZEPOSITION = enum {
	M_EMPTY,
	M_LADDER_UP,
	M_LADDER_DOWN,
	M_SPECIAL
    },

    MAZEWALL = enum {
	W_WALL,
	W_EMPTY,
	W_DOOR,
	W_SPECIAL
    },

    SCREENLINE = unsigned SCREENLINES - 1,
    SCREENCOL = unsigned SCREENCOLS - 1,

    PATTERNLINE = unsigned PATTERNLINES - 1,
    PATTERNCOL = unsigned PATTERNCOLS - 1,
    PATTERNMLINE = unsigned PATTERNMLINES - 1,
    PATTERNMCOL = unsigned PATTERNMCOLS - 1,

    DISPLAYLINE = unsigned DISPLAYLINES - 1,
    DISPLAYCOL = unsigned DISPLAYCOLS - 1,

    GRIDLINE = unsigned 255,
    GRIDCOL = unsigned 255,

    MAZELINE = unsigned 255,
    MAZECOL = unsigned 255,

    GRID = unsigned GRIDMAX,
    MAZE = unsigned MAZEMAX,
    PATTERN = unsigned PATTERNMAX + 1,
    PICTURE = unsigned PICTUREMAX,
    POLYGON = unsigned POLYGONMAX,
    VERTEX = unsigned VERTEXMAX,
    GRIDSPECIAL = unsigned GRIDSPECIALMAX,
    BITPOS = unsigned BITPOSMAX,
    OBJECT = unsigned OBJECTMAX,
    MONSTER = unsigned MONSTERMAX,
    TERRITORY = unsigned TERRITORYMAX,
    FULLPALETTE = unsigned FULLPALETTEMAX,
    HALFPALETTE = unsigned HALFPALETTEMAX,

    TERRAIN = unsigned 1 << TERRAINBITS - 1,

/* structure of a 16 x 16 pattern for use on the screen: */

    PATTERNTYPE = struct {
	[PATTERNDEPTH, PATTERNMLINES, PATTERNMCOLS] uint pa_data;
	bool pa_topRound, pa_rightRound, pa_bottomRound, pa_leftRound;
	ushort pa_group;
    },

/* structure of a mazewalk picture of a monster, etc.: */

    PICTURETYPE = struct {
	POLYGON p_start;
	ushort p_polyCount;
    },

/* structure of a polygon for use in pictures: */

    POLYGONTYPE = struct {
	VERTEX pg_start;
	ushort pg_vertexCount;
	ushort pg_fgColour, pg_bgColour, pg_pattern;
    },

/* structure of a vertex for use in picture polygons: */

    VERTEXTYPE = struct {
	ushort v_line;
	ushort v_column;
    },

/* structure of an object type: */

    OBJECTTYPE = struct {
	uint ob_name;
	ATTACK ob_type;
	ushort ob_strength;
	ushort ob_weight;
	ushort ob_weakness;
    },

/* structure of a monster type: */

    MONSTERTYPE = struct {
	uint m_name;
	PATTERN m_pattern;
	PICTURE m_picture;
	HALFPALETTE m_palette;
	uint m_image;
	ushort m_strength;
	ushort m_hitPoints;
	ushort m_speed;
	ATTACK m_type;
	[range(ATTACK) - 1] short m_vulnerability;
    },

/* structure of an occupant monster type: */

    OCCUPANTTYPE = struct {
	MONSTER oc_type;
	ushort oc_probability;
    },

/* structure of a territory type: */

    TERRITORYTYPE = struct {
	PATTERN t_pattern;
	[OCCUPANTCOUNT] OCCUPANTTYPE t_occupants;
    },

/* structure of a special display element: */

    GRIDSPECIALTYPE = struct {
	GRID gsp_grid;
	GRIDLINE gsp_line;
	GRIDCOL gsp_col;
	TERRITORY gsp_territory;
    },

/* structure of a grid: */

    GRIDTYPE = struct {
	GRIDLINE g_lines;
	GRIDCOL g_cols;
	[range(TERRAIN)] TERRITORY g_territorys;
	BITPOS g_start;
	FULLPALETTE g_palette;
    },

/* structure of a maze: */

    MAZETYPE = struct {
	MAZELINE ma_lines;
	MAZECOL ma_cols;
	BITPOS ma_start;
	[OCCUPANTCOUNT] OCCUPANTTYPE ma_occupants;
	HALFPALETTE ma_palette;
    },

    CYCLETYPE = struct {
	ushort cy_firstColour, cy_lastColour, cy_period;
    },

    FULLPALETTEARRAY = [PATTERNCOLOURS] uint,
    HALFPALETTEARRAY = [PICTURECOLOURS] uint,

    FULLPALETTETYPE = struct {
	FULLPALETTEARRAY fp_palette;
	[4] CYCLETYPE fp_cycles;
    },

    HALFPALETTETYPE = struct {
	HALFPALETTEARRAY hp_palette;
	[2] CYCLETYPE hp_cycles;
    };

TERRAIN T_SPECIAL = range(TERRAIN);
PATTERN NOPATTERN = range(PATTERN);
