/*
 * user.g - include file for game-writer interface to Explore.
 *	Requires previous inclusion of 'explore.g'
 */

uint
    MAXCARRY = 25,
    DROPPEDMAX = 50,
    NP = 53 + 17;		/* number of proc fields in Vector_t */

type
    DROPPED = struct {
	bool d_onGrid;
	ushort d_gridMaze;
	ushort d_line, d_col;
	OBJECT d_object;
    },

    HEADING = enum {NORTH, EAST, SOUTH, WEST},

    Vector_t = struct {

    /* entries filled in by Explore */

	/* I/O routines */

	proc(char ch)void PutChar;
	proc(*char mess)void PMess;
	proc()void Pnl;
	proc(*char mess)void Print;
	proc(uint n)void PNumber;
	proc(uint recno)void PText;
	proc()char ReadChar;
	proc(*char prompt, buff; ushort len)void ReadChars;
	proc(*char prompt)uint ReadNum;
	proc()void Beep;

	/* general utility routines */

	proc(uint rang)uint Random;
	proc(*char str)uint CharsLen;
	proc(char ch)char Cap;

	/* general graphics routines */

	proc()void GraphicsDown;
	proc()void ToggleGraphics;
	proc(ushort l, c, colour)void Point;
	proc(ushort l1, c1, l2, c2; bool xor; ushort colour)void Line;
	proc(ushort l1, c1, l2, c2; bool xor; ushort colour)void Rectangle;
	proc(PICTURE p)void PicDisp;
	proc(FULLPALETTE fp)void SetFullPalette;
	proc(HALFPALETTE hp)void SetBottomPalette;
	proc(HALFPALETTE hp)void SetTopPalette;
	proc(uint which, colour)void SetColour;
	proc(uint which)void DisplayImage;	

	/* utilities for accessing data structures */

	proc(MONSTER m)*MONSTERTYPE GetMonster;
	proc(ATTACK a)void PAttType;
	proc(uint pos)*char GetName;
	proc(OBJECT o)*OBJECTTYPE GetObject;

	/* object manipulation routines */

	proc(OBJECT o)void Carry;
	proc(OBJECT o)void Drop;
	proc(OBJECT o; bool onGr; ushort gridMaze, line, column)void Place;
	proc(OBJECT o; bool onGr; ushort gridMaze, line, column)void Remove;
	proc(OBJECT o)void PlaceHere;
	proc(OBJECT o)bool IsHere;
	proc(OBJECT o)bool IsCarrying;
	proc()void StuffCheck;
	proc()void HereList;

	/* fighting routines */

	proc()void MAtt;
	proc()void SpecificMAtt;
	proc()void MHit;
	proc(ATTACK typ; uint strength; *char missMess, hitMess)void PHit;
	proc()void RemoveMonster;

	/* routines for grid things */

	proc()void GridSetup;
	proc()void GridDisplay;
	proc(char ch)void GridStandard;
	proc(PATTERN p)void SetNewCursor;
	proc(ushort line, column; *PATTERNTYPE pP)void SaveBackGround;
	proc(ushort line, column; PATTERN p)void DisplayMonster;
	proc(ushort line, column; *PATTERNTYPE pP)void RestoreBackGround;

	/* routines for maze things */

	proc()void MazeSetup;
	proc()void MazeDisplay;
	proc(char ch)void MazeStandard;
	proc(MAZELINE l; MAZECOL c)ushort GConf;

    /* entries to be filled in by the scenario */

	/* general purpose stuff */

	proc()void Restore;
	proc()bool Save;
	proc(bool restoring)void Run;
	proc()void Stats;

	/* object stuff */

	proc(OBJECT o)bool Dropped;
	proc(OBJECT o)void Use;

	/* fighting stuff */

	proc()void MonsterCheck;
	proc()void KillMonster;

	/* grid stuff */

	proc(GRIDLINE line;GRIDCOL col;TERRITORY t;ushort group)bool GridNorm;
	proc(GRIDLINE line;GRIDCOL col;TERRITORY t;ushort group)bool GridSpec;
	proc()void GridMoved;

	/* maze stuff */

	proc(MAZELINE line; MAZECOL col)bool MazeMove;
	proc()void MazeUp;
	proc()void MazeDown;
	proc()void MazeMoved;
	proc()void MazeSpec;
	proc(MAZELINE line; MAZECOL col; bool isHor)bool MazeSpecDoor;

    /* special variables used for intercommunication: */

	bool gameOver;

	bool onGrid;
	GRID grid;
	MAZE maze;
	ushort maxView;
	PATTERN cursorPattern;
	HEADING heading;
	TERRITORY gridTerritory;

	GRIDLINE
	    gridLines,
	    windowLineTop,
	    gridLine;
	GRIDCOL
	    gridCols,
	    windowColLeft,
	    gridCol;
	MAZELINE
	    mazeLines,
	    mazeLine;
	MAZECOL
	    mazeCols,
	    mazeCol;

	uint hitPoints, strength, experience, money, food, health;
	[range(ATTACK) - 1] uint protection;

	uint maxCarryWeight;
	ushort maxCarryCount;
	uint carryWeight;
	ushort carryCount;
	[MAXCARRY] OBJECT carrying;

	ushort droppedCount;
	[DROPPEDMAX] DROPPED droppedList;

	bool armed;
	OBJECT weaponNumber;

	bool underAttack;
	MONSTER monsterNumber;
	MONSTERTYPE monsterData;

	bool printing;
    };

*Vector_t V;
